from .api_client import OctomilClientError
from .auth import DeviceAuthClient
from .control_plane import (
    ExperimentsAPI,
    FederatedAnalyticsAPI,
    FederatedAnalyticsClient,
    RolloutsAPI,
)
from .edge import Octomil
from .federated_client import FederatedClient, apply_filters, compute_state_dict_delta
from .federation import Federation
from .filters import DataKind, DeltaFilter, FilterRegistry, FilterResult
from .registry import ModelRegistry
from .secagg import (
    HKDF_INFO_PAIRWISE_MASK,
    HKDF_INFO_SELF_MASK,
    HKDF_INFO_SHARE_ENCRYPTION,
    SECAGG_PLUS_MOD_RANGE,
    ECKeyPair,
    SecAggClient,
    SecAggConfig,
    SecAggPlusClient,
    SecAggPlusConfig,
)

__all__ = [
    "Octomil",
    "OctomilClientError",
    "Federation",
    "FederatedClient",
    "ModelRegistry",
    "RolloutsAPI",
    "ExperimentsAPI",
    "FederatedAnalyticsClient",
    "FederatedAnalyticsAPI",
    "compute_state_dict_delta",
    "apply_filters",
    "DataKind",
    "DeltaFilter",
    "FilterRegistry",
    "FilterResult",
    "DeviceAuthClient",
    "ECKeyPair",
    "SecAggClient",
    "SecAggConfig",
    "SecAggPlusClient",
    "SecAggPlusConfig",
    "SECAGG_PLUS_MOD_RANGE",
    "HKDF_INFO_PAIRWISE_MASK",
    "HKDF_INFO_SHARE_ENCRYPTION",
    "HKDF_INFO_SELF_MASK",
]



import asyncio
import unittest
from datetime import datetime, timedelta, timezone
from unittest.mock import patch

from octomil.auth import DeviceAuthClient, DeviceTokenState


class _FakeKeyring:
    def __init__(self):
        self.store = {}

    def set_password(self, service, key, value):
        self.store[(service, key)] = value

    def get_password(self, service, key):
        return self.store.get((service, key))

    def delete_password(self, service, key):
        self.store.pop((service, key), None)


class _FakeResponse:
    def __init__(self, status_code: int, payload: dict):
        self.status_code = status_code
        self._payload = payload

    def json(self):
        return self._payload

    def raise_for_status(self):
        if self.status_code >= 400:
            raise RuntimeError(f"http {self.status_code}")


class _FakeAsyncClient:
    requests_log = []

    def __init__(self, *args, **kwargs):
        self._requests = []

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        return False

    async def post(self, url, json=None, headers=None):
        await asyncio.sleep(0)  # yield to event loop
        self._requests.append((url, json, headers))
        self.__class__.requests_log.append((url, json, headers))
        if url.endswith("/bootstrap"):
            payload = {
                "access_token": "acc_bootstrap",
                "refresh_token": "ref_bootstrap",
                "token_type": "Bearer",
                "expires_in": 900,
                "org_id": "org_1",
                "device_identifier": "device_1",
                "scopes": ["devices:write"],
            }
            return _FakeResponse(200, payload)
        if url.endswith("/refresh"):
            payload = {
                "access_token": "acc_refresh",
                "refresh_token": "ref_refresh",
                "token_type": "Bearer",
                "expires_in": 900,
                "org_id": "org_1",
                "device_identifier": "device_1",
                "scopes": ["devices:write"],
            }
            return _FakeResponse(200, payload)
        if url.endswith("/revoke"):
            return _FakeResponse(200, {})
        return _FakeResponse(404, {})


class _FailRefreshAsyncClient(_FakeAsyncClient):
    async def post(self, url, json=None, headers=None):
        if url.endswith("/refresh"):
            raise RuntimeError("network down")
        return await super().post(url, json=json, headers=headers)


class DeviceAuthClientTests(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        _FakeAsyncClient.requests_log = []

    async def test_bootstrap_refresh_revoke_lifecycle(self):
        fake_keyring = _FakeKeyring()
        with patch("octomil.auth.keyring", fake_keyring), patch(
            "octomil.auth.httpx.AsyncClient", _FakeAsyncClient
        ):
            client = DeviceAuthClient(
                base_url="https://api.example.com",
                org_id="org_1",
                device_identifier="device_1",
            )
            state = await client.bootstrap(bootstrap_bearer_token="bootstrap_token")
            self.assertEqual(state.access_token, "acc_bootstrap")

            refreshed = await client.refresh()
            self.assertEqual(refreshed.access_token, "acc_refresh")

            await client.revoke()
            self.assertIsNone(client._load_token_state())

    async def test_bootstrap_sends_expected_payload_and_bearer_token(self):
        fake_keyring = _FakeKeyring()
        with patch("octomil.auth.keyring", fake_keyring), patch(
            "octomil.auth.httpx.AsyncClient", _FakeAsyncClient
        ):
            client = DeviceAuthClient(
                base_url="https://api.example.com",
                org_id="org_1",
                device_identifier="device_1",
            )
            await client.bootstrap(
                bootstrap_bearer_token="bootstrap_token",
                scopes=["devices:write", "heartbeat:write"],
                access_ttl_seconds=600,
                device_id="device_db_id",
            )

            bootstrap_url, bootstrap_payload, bootstrap_headers = (
                _FakeAsyncClient.requests_log[0]
            )
            self.assertTrue(bootstrap_url.endswith("/api/v1/device-auth/bootstrap"))
            self.assertEqual(
                bootstrap_headers["Authorization"], "Bearer bootstrap_token"
            )
            self.assertEqual(bootstrap_payload["org_id"], "org_1")
            self.assertEqual(bootstrap_payload["device_identifier"], "device_1")
            self.assertEqual(bootstrap_payload["access_ttl_seconds"], 600)
            self.assertEqual(bootstrap_payload["device_id"], "device_db_id")
            self.assertEqual(
                bootstrap_payload["scopes"], ["devices:write", "heartbeat:write"]
            )

    async def test_refresh_uses_rotated_refresh_token(self):
        fake_keyring = _FakeKeyring()
        with patch("octomil.auth.keyring", fake_keyring), patch(
            "octomil.auth.httpx.AsyncClient", _FakeAsyncClient
        ):
            client = DeviceAuthClient(
                base_url="https://api.example.com",
                org_id="org_1",
                device_identifier="device_1",
            )
            await client.bootstrap(bootstrap_bearer_token="bootstrap_token")
            await client.refresh()
            await client.refresh()

            refresh_requests = [
                payload
                for url, payload, _headers in _FakeAsyncClient.requests_log
                if url.endswith("/api/v1/device-auth/refresh")
            ]
            self.assertEqual(refresh_requests[0]["refresh_token"], "ref_bootstrap")
            self.assertEqual(refresh_requests[1]["refresh_token"], "ref_refresh")

    async def test_get_access_token_offline_fallback_before_expiry(self):
        fake_keyring = _FakeKeyring()
        with patch("octomil.auth.keyring", fake_keyring), patch(
            "octomil.auth.httpx.AsyncClient", _FailRefreshAsyncClient
        ):
            client = DeviceAuthClient(
                base_url="https://api.example.com",
                org_id="org_1",
                device_identifier="device_1",
            )
            state = DeviceTokenState(
                access_token="still_valid",
                refresh_token="refresh_token",
                token_type="Bearer",
                expires_at=datetime.now(timezone.utc) + timedelta(minutes=5),
                org_id="org_1",
                device_identifier="device_1",
                scopes=["devices:write"],
            )
            client._store_token_state(state)

            token = await client.get_access_token(
                refresh_if_expiring_within_seconds=600
            )
            self.assertEqual(token, "still_valid")

    async def test_get_access_token_raises_if_expired_and_refresh_fails(self):
        fake_keyring = _FakeKeyring()
        with patch("octomil.auth.keyring", fake_keyring), patch(
            "octomil.auth.httpx.AsyncClient", _FailRefreshAsyncClient
        ):
            client = DeviceAuthClient(
                base_url="https://api.example.com",
                org_id="org_1",
                device_identifier="device_1",
            )
            state = DeviceTokenState(
                access_token="expired",
                refresh_token="refresh_token",
                token_type="Bearer",
                expires_at=datetime.now(timezone.utc) - timedelta(seconds=1),
                org_id="org_1",
                device_identifier="device_1",
                scopes=["devices:write"],
            )
            client._store_token_state(state)

            with self.assertRaises(RuntimeError):
                await client.get_access_token(refresh_if_expiring_within_seconds=30)

    async def test_revoke_failure_preserves_stored_state(self):
        fake_keyring = _FakeKeyring()

        class _FailRevokeAsyncClient(_FakeAsyncClient):
            async def post(self, url, json=None, headers=None):
                if url.endswith("/revoke"):
                    raise RuntimeError("network down")
                return await super().post(url, json=json, headers=headers)

        with patch("octomil.auth.keyring", fake_keyring), patch(
            "octomil.auth.httpx.AsyncClient", _FailRevokeAsyncClient
        ):
            client = DeviceAuthClient(
                base_url="https://api.example.com",
                org_id="org_1",
                device_identifier="device_1",
            )
            await client.bootstrap(bootstrap_bearer_token="bootstrap_token")

            with self.assertRaises(RuntimeError):
                await client.revoke()
            self.assertIsNotNone(client._load_token_state())

    def test_device_token_state_from_response_with_expires_at(self):
        payload = {
            "access_token": "acc_token",
            "refresh_token": "ref_token",
            "token_type": "Bearer",
            "expires_at": "2024-12-31T23:59:59Z",
            "org_id": "org_1",
            "device_identifier": "device_1",
            "scopes": ["devices:write"],
        }
        state = DeviceTokenState.from_response(payload)
        self.assertEqual(state.access_token, "acc_token")
        self.assertEqual(state.expires_at.year, 2024)

    def test_device_token_state_from_response_with_expires_in(self):
        payload = {
            "access_token": "acc_token",
            "refresh_token": "ref_token",
            "expires_in": 900,
            "org_id": "org_1",
            "device_identifier": "device_1",
        }
        state = DeviceTokenState.from_response(payload)
        self.assertEqual(state.access_token, "acc_token")
        self.assertEqual(state.token_type, "Bearer")
        self.assertEqual(state.scopes, [])

    def test_device_token_state_serialization(self):
        state = DeviceTokenState(
            access_token="acc",
            refresh_token="ref",
            token_type="Bearer",
            expires_at=datetime.now(timezone.utc) + timedelta(minutes=15),
            org_id="org_1",
            device_identifier="device_1",
            scopes=["devices:write"],
        )
        json_str = state.to_json()
        restored = DeviceTokenState.from_json(json_str)
        self.assertEqual(restored.access_token, state.access_token)
        self.assertEqual(restored.refresh_token, state.refresh_token)
        self.assertEqual(restored.org_id, state.org_id)

    async def test_clear_token_state(self):
        fake_keyring = _FakeKeyring()
        with patch("octomil.auth.keyring", fake_keyring), patch(
            "octomil.auth.httpx.AsyncClient", _FakeAsyncClient
        ):
            client = DeviceAuthClient(
                base_url="https://api.example.com",
                org_id="org_1",
                device_identifier="device_1",
            )
            await client.bootstrap(bootstrap_bearer_token="bootstrap_token")
            self.assertIsNotNone(client._load_token_state())
            client.clear_token_state()
            self.assertIsNone(client._load_token_state())

    def test_get_access_token_sync_outside_loop(self):
        """Test get_access_token_sync works when called outside an event loop"""
        fake_keyring = _FakeKeyring()
        with patch("octomil.auth.keyring", fake_keyring), patch(
            "octomil.auth.httpx.AsyncClient", _FakeAsyncClient
        ):
            client = DeviceAuthClient(
                base_url="https://api.example.com",
                org_id="org_1",
                device_identifier="device_1",
            )
            state = DeviceTokenState(
                access_token="valid_token",
                refresh_token="refresh_token",
                token_type="Bearer",
                expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
                org_id="org_1",
                device_identifier="device_1",
                scopes=["devices:write"],
            )
            client._store_token_state(state)
            # This is a non-async test method, so no event loop is running
            token = client.get_access_token_sync()
            self.assertEqual(token, "valid_token")

    async def test_get_access_token_sync_raises_inside_loop(self):
        """Test get_access_token_sync raises when called inside an event loop"""
        await asyncio.sleep(0)  # ensure event loop is active
        fake_keyring = _FakeKeyring()
        with patch("octomil.auth.keyring", fake_keyring):
            client = DeviceAuthClient(
                base_url="https://api.example.com",
                org_id="org_1",
                device_identifier="device_1",
            )
            with self.assertRaises(RuntimeError) as ctx:
                client.get_access_token_sync()
            self.assertIn(
                "cannot be called inside an active event loop", str(ctx.exception)
            )

    async def test_get_access_token_no_token_state(self):
        fake_keyring = _FakeKeyring()
        with patch("octomil.auth.keyring", fake_keyring):
            client = DeviceAuthClient(
                base_url="https://api.example.com",
                org_id="org_1",
                device_identifier="device_1",
            )
            with self.assertRaises(RuntimeError) as ctx:
                await client.get_access_token()
            self.assertIn("No token state found", str(ctx.exception))

    async def test_refresh_no_token_state(self):
        fake_keyring = _FakeKeyring()
        with patch("octomil.auth.keyring", fake_keyring):
            client = DeviceAuthClient(
                base_url="https://api.example.com",
                org_id="org_1",
                device_identifier="device_1",
            )
            with self.assertRaises(RuntimeError) as ctx:
                await client.refresh()
            self.assertIn("No token state found", str(ctx.exception))

    async def test_revoke_with_no_state_is_noop(self):
        fake_keyring = _FakeKeyring()
        with patch("octomil.auth.keyring", fake_keyring), patch(
            "octomil.auth.httpx.AsyncClient", _FakeAsyncClient
        ):
            client = DeviceAuthClient(
                base_url="https://api.example.com",
                org_id="org_1",
                device_identifier="device_1",
            )
            await client.revoke()


if __name__ == "__main__":
    unittest.main()



"""Example tests demonstrating non-IID data partitioning with flwr-datasets.

These tests show how ``flwr-datasets`` can be combined with the Octomil SDK
to simulate realistic heterogeneous federated learning scenarios.

Requirements:
    pip install "octomil-sdk[dev]"  # includes flwr-datasets
"""

import unittest

try:
    from flwr_datasets import FederatedDataset
    from flwr_datasets.partitioner import (
        DirichletPartitioner,
        ShardPartitioner,
    )

    HAS_FLWR_DATASETS = True
except ImportError:
    HAS_FLWR_DATASETS = False


@unittest.skipUnless(HAS_FLWR_DATASETS, "flwr-datasets not installed")
class NonIIDPartitioningTests(unittest.TestCase):
    """Demonstrate non-IID partitioning strategies relevant to FL."""

    def test_dirichlet_partitioning_cifar10(self):
        """Dirichlet(alpha) partitioning creates label-skewed splits.

        A low alpha (e.g. 0.1) makes each partition dominated by a few
        classes, which is the classic non-IID scenario in FL literature.
        """
        fds = FederatedDataset(
            dataset="cifar10",
            partitioners={
                "train": DirichletPartitioner(
                    num_partitions=5,
                    partition_by="label",
                    alpha=0.5,
                ),
            },
        )

        partition_0 = fds.load_partition(0)
        self.assertGreater(len(partition_0), 0)

        # Each partition should have data but likely not all 10 labels equally.
        labels = set(partition_0["label"])
        self.assertGreater(len(labels), 0)

    def test_shard_partitioning_mnist(self):
        """Shard partitioning assigns fixed-size, label-sorted shards to clients.

        This replicates the original FedAvg paper's non-IID setup where each
        client gets 2 shards of sorted data (so most clients see only 2 digits).
        """
        fds = FederatedDataset(
            dataset="mnist",
            partitioners={
                "train": ShardPartitioner(
                    num_partitions=10,
                    partition_by="label",
                    num_shards_per_partition=2,
                ),
            },
        )

        partition_3 = fds.load_partition(3)
        self.assertGreater(len(partition_3), 0)

        # With 2 shards per partition the label diversity should be limited.
        labels = set(partition_3["label"])
        self.assertLessEqual(len(labels), 4)  # generous upper bound

    def test_multiple_partitions_cover_full_dataset(self):
        """All partitions together should contain every sample exactly once."""
        num_partitions = 5
        fds = FederatedDataset(
            dataset="cifar10",
            partitioners={
                "train": DirichletPartitioner(
                    num_partitions=num_partitions,
                    partition_by="label",
                    alpha=1.0,
                ),
            },
        )

        total_samples = sum(len(fds.load_partition(i)) for i in range(num_partitions))
        # CIFAR-10 train set has 50 000 samples.
        self.assertEqual(total_samples, 50_000)

    def test_partition_to_pandas(self):
        """Partitions can be converted to pandas DataFrames for use with the SDK."""
        fds = FederatedDataset(
            dataset="mnist",
            partitioners={
                "train": DirichletPartitioner(
                    num_partitions=3,
                    partition_by="label",
                    alpha=0.5,
                ),
            },
        )

        partition = fds.load_partition(0)
        df = partition.to_pandas()
        self.assertIn("label", df.columns)
        self.assertGreater(len(df), 0)


if __name__ == "__main__":
    unittest.main()



"""Managed venv + engine setup for the PyInstaller binary.

After ``curl -fsSL https://get.octomil.com | sh``, a background process
calls ``octomil setup --foreground`` which:

1. Finds a system Python with venv support.
2. Creates ``~/.octomil/engines/venv/``.
3. Installs ``octomil-sdk[mlx,serve]`` (Apple Silicon) or
   ``octomil-sdk[llama,serve]`` (elsewhere).
4. Downloads the recommended model via huggingface_hub.

When ``octomil serve`` runs from the frozen binary and finds no native
engines, it calls ``get_venv_python()`` + ``is_engine_ready()`` and
re-execs into the venv's Python via ``os.execv()``.
"""

from __future__ import annotations

import json
import logging
import os
import platform
import shutil
import subprocess
import time
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Optional

logger = logging.getLogger(__name__)

# ---------------------------------------------------------------------------
# Paths
# ---------------------------------------------------------------------------

OCTOMIL_DIR = Path.home() / ".octomil"
ENGINES_DIR = OCTOMIL_DIR / "engines"
VENV_DIR = ENGINES_DIR / "venv"
STATE_FILE = OCTOMIL_DIR / "setup_state.json"
SETUP_LOG = OCTOMIL_DIR / "setup.log"

# ---------------------------------------------------------------------------
# State tracking
# ---------------------------------------------------------------------------

PHASE_PENDING = "pending"
PHASE_CREATING_VENV = "creating_venv"
PHASE_INSTALLING_ENGINE = "installing_engine"
PHASE_DOWNLOADING_MODEL = "downloading_model"
PHASE_COMPLETE = "complete"
PHASE_FAILED = "failed"


@dataclass
class SetupState:
    """Persistent state for the setup pipeline."""

    phase: str = PHASE_PENDING
    engine: Optional[str] = None
    package: Optional[str] = None
    engine_installed: bool = False
    model_key: Optional[str] = None
    model_downloaded: bool = False
    error: Optional[str] = None
    started_at: Optional[float] = None
    finished_at: Optional[float] = None


def load_state() -> SetupState:
    """Load setup state from disk."""
    try:
        data = json.loads(STATE_FILE.read_text())
        return SetupState(**{k: v for k, v in data.items() if k in SetupState.__dataclass_fields__})
    except (FileNotFoundError, json.JSONDecodeError, TypeError):
        return SetupState()


def _save_state(state: SetupState) -> None:
    """Persist setup state to disk."""
    STATE_FILE.parent.mkdir(parents=True, exist_ok=True)
    STATE_FILE.write_text(json.dumps(asdict(state), indent=2) + "\n")


# ---------------------------------------------------------------------------
# Public queries
# ---------------------------------------------------------------------------


def get_venv_python() -> Optional[str]:
    """Return path to the managed venv's Python, or None."""
    p = VENV_DIR / "bin" / "python"
    if p.is_file():
        return str(p)
    return None


def is_engine_ready() -> bool:
    """True if the managed venv has an engine installed."""
    state = load_state()
    return state.engine_installed and state.phase == PHASE_COMPLETE


def is_setup_in_progress() -> bool:
    """True if setup is currently running (not finished or failed)."""
    state = load_state()
    return state.phase not in (PHASE_PENDING, PHASE_COMPLETE, PHASE_FAILED)


def is_setup_complete() -> bool:
    """True if setup finished successfully."""
    return load_state().phase == PHASE_COMPLETE


# ---------------------------------------------------------------------------
# System Python detection
# ---------------------------------------------------------------------------

_PYTHON_CANDIDATES = [
    "python3.13",
    "python3.12",
    "python3.11",
    "python3.10",
    "python3.9",
    "python3",
]


def _is_correct_arch(python_path: str) -> bool:
    """Verify the Python binary matches the current architecture.

    On Apple Silicon Macs running Rosetta, a system python3 might be
    x86_64 which would produce an x86_64 venv — no good for mlx-lm.
    """
    machine = platform.machine()
    if machine != "arm64":
        return True  # Only matters on Apple Silicon

    try:
        # Check the binary itself
        arch_result = subprocess.run(
            ["file", python_path],
            capture_output=True,
            text=True,
            timeout=5,
        )
        if "x86_64" in arch_result.stdout and "arm64" not in arch_result.stdout:
            return False
    except Exception:
        pass
    return True


def _has_venv_module(python_path: str) -> bool:
    """Check if the Python binary has the venv module available."""
    try:
        result = subprocess.run(
            [python_path, "-c", "import venv"],
            capture_output=True,
            timeout=10,
        )
        return result.returncode == 0
    except Exception:
        return False


def find_system_python() -> Optional[str]:
    """Find a suitable system Python for creating the managed venv.

    Searches python3.13 down to python3.9 on PATH, verifies:
    - venv module is available
    - architecture matches (arm64 on Apple Silicon)

    Returns the full path or None.
    """
    for name in _PYTHON_CANDIDATES:
        path = shutil.which(name)
        if path is None:
            continue
        if not _has_venv_module(path):
            logger.debug("Skipping %s: no venv module", path)
            continue
        if not _is_correct_arch(path):
            logger.debug("Skipping %s: wrong architecture", path)
            continue
        logger.debug("Found system Python: %s", path)
        return path

    return None


# ---------------------------------------------------------------------------
# Engine detection
# ---------------------------------------------------------------------------


def detect_best_engine() -> tuple[str, str]:
    """Detect the best engine for this platform.

    Returns:
        (engine_name, pip_package) — e.g. ("mlx-lm", "octomil-sdk[mlx,serve]")
    """
    system = platform.system()
    machine = platform.machine()

    if system == "Darwin" and machine == "arm64":
        return ("mlx-lm", "octomil-sdk[mlx,serve]")
    else:
        return ("llama.cpp", "octomil-sdk[llama,serve]")


# ---------------------------------------------------------------------------
# Recommended model
# ---------------------------------------------------------------------------


def _get_recommended_model() -> str:
    """Pick the recommended model key for first-time download.

    Uses the same logic as the launcher's auto-select but simplified:
    pick the best model that fits the device memory budget.
    """
    try:
        from octomil.agents.launcher import _build_recommendations

        recommendations = _build_recommendations()
        if recommendations:
            return recommendations[0].key
    except Exception:
        pass
    # Fallback: safe default that runs everywhere
    return "qwen-coder-7b"


# ---------------------------------------------------------------------------
# Venv creation
# ---------------------------------------------------------------------------


def _has_uv() -> bool:
    """Check if uv is on PATH."""
    return shutil.which("uv") is not None


def create_managed_venv(python_path: str) -> str:
    """Create the managed venv at ~/.octomil/engines/venv/.

    Prefers ``uv venv`` for speed, falls back to ``python3 -m venv``.

    Returns the path to the venv's python binary.
    """
    VENV_DIR.parent.mkdir(parents=True, exist_ok=True)

    # Remove existing venv if present
    if VENV_DIR.exists():
        shutil.rmtree(VENV_DIR)

    if _has_uv():
        logger.info("Creating venv with uv...")
        subprocess.run(
            ["uv", "venv", str(VENV_DIR), "--python", python_path],
            check=True,
            capture_output=True,
        )
    else:
        logger.info("Creating venv with python -m venv...")
        subprocess.run(
            [python_path, "-m", "venv", str(VENV_DIR)],
            check=True,
            capture_output=True,
        )

    venv_python = str(VENV_DIR / "bin" / "python")
    if not os.path.isfile(venv_python):
        raise RuntimeError(f"Venv created but python not found at {venv_python}")

    return venv_python


# ---------------------------------------------------------------------------
# Engine installation
# ---------------------------------------------------------------------------


def install_engine(package: str) -> None:
    """Install the engine package into the managed venv.

    Prefers ``uv pip install`` for speed, falls back to venv pip.
    """
    venv_python = str(VENV_DIR / "bin" / "python")

    if _has_uv():
        logger.info("Installing %s with uv...", package)
        subprocess.run(
            ["uv", "pip", "install", package, "--python", venv_python],
            check=True,
        )
    else:
        logger.info("Installing %s with pip...", package)
        subprocess.run(
            [venv_python, "-m", "pip", "install", "--upgrade", "pip"],
            check=True,
            capture_output=True,
        )
        subprocess.run(
            [venv_python, "-m", "pip", "install", package],
            check=True,
        )


# ---------------------------------------------------------------------------
# Model download
# ---------------------------------------------------------------------------


def download_model(model_key: str) -> None:
    """Download a model using the venv's huggingface_hub.

    Runs ``python -c 'from huggingface_hub import snapshot_download; ...'``
    inside the managed venv so we don't need huggingface_hub in the
    frozen binary.
    """
    venv_python = get_venv_python()
    if venv_python is None:
        raise RuntimeError("Managed venv not found — run setup first")

    # Resolve the model key to an HF repo via the venv's octomil
    # This runs the catalog resolution in the venv where all deps exist
    script = f"""\
import sys
try:
    from octomil.models.resolver import resolve
    result = resolve("{model_key}")
    repo = result.hf_repo
    filename = result.filename
except Exception as e:
    # Fallback: try catalog directly
    try:
        from octomil.models.catalog import get_model
        entry = get_model("{model_key}")
        if entry is None:
            print("ERROR:Model not found in catalog", file=sys.stderr)
            sys.exit(1)
        variant = entry.variants.get(entry.default_quant)
        if variant is None:
            print("ERROR:No default variant", file=sys.stderr)
            sys.exit(1)
        # Prefer MLX on Apple Silicon
        import platform
        if platform.machine() == "arm64" and variant.mlx:
            repo = variant.mlx
            filename = None
        elif variant.gguf:
            repo = variant.gguf.repo
            filename = variant.gguf.filename
        else:
            print(f"ERROR:No downloadable artifact for {{entry.default_quant}}", file=sys.stderr)
            sys.exit(1)
    except Exception as e2:
        print(f"ERROR:{{e2}}", file=sys.stderr)
        sys.exit(1)

if filename:
    from huggingface_hub import hf_hub_download
    path = hf_hub_download(repo, filename)
    print(f"Downloaded {{filename}} from {{repo}} -> {{path}}")
else:
    from huggingface_hub import snapshot_download
    path = snapshot_download(repo)
    print(f"Downloaded {{repo}} -> {{path}}")
"""
    result = subprocess.run(
        [venv_python, "-c", script],
        capture_output=True,
        text=True,
    )

    if result.returncode != 0:
        stderr = result.stderr.strip()
        raise RuntimeError(f"Model download failed: {stderr}")

    logger.info("Model download output: %s", result.stdout.strip())


# ---------------------------------------------------------------------------
# Full setup pipeline
# ---------------------------------------------------------------------------


def run_setup(*, force: bool = False, foreground: bool = False) -> SetupState:
    """Run the full setup pipeline.

    Args:
        force: Re-run even if already complete.
        foreground: If True, suppresses interactive output (used by install.sh).

    Returns:
        Final SetupState.
    """
    state = load_state()

    # Skip if already complete (unless forced)
    if state.phase == PHASE_COMPLETE and not force:
        return state

    def _log(msg: str) -> None:
        logger.info(msg)
        if not foreground:
            import click

            click.echo(msg)

    state = SetupState(started_at=time.time())
    _save_state(state)

    # Step 1: Find system Python
    _log("Finding system Python...")
    python_path = find_system_python()
    if python_path is None:
        state.phase = PHASE_FAILED
        state.error = (
            "No suitable Python found. Install python3 with venv support:\n"
            "  macOS: brew install python3\n"
            "  Ubuntu/Debian: sudo apt install python3-venv\n"
            "  Fedora: sudo dnf install python3"
        )
        state.finished_at = time.time()
        _save_state(state)
        _log(f"Setup failed: {state.error}")
        return state

    _log(f"  Using {python_path}")

    # Step 2: Detect best engine
    engine_name, package = detect_best_engine()
    state.engine = engine_name
    state.package = package
    _save_state(state)
    _log(f"  Engine: {engine_name} ({package})")

    # Step 3: Create venv
    state.phase = PHASE_CREATING_VENV
    _save_state(state)
    _log("Creating managed venv...")
    try:
        create_managed_venv(python_path)
        _log(f"  Venv: {VENV_DIR}")
    except Exception as e:
        state.phase = PHASE_FAILED
        state.error = f"Failed to create venv: {e}"
        state.finished_at = time.time()
        _save_state(state)
        _log(f"Setup failed: {state.error}")
        return state

    # Step 4: Install engine
    state.phase = PHASE_INSTALLING_ENGINE
    _save_state(state)
    _log(f"Installing {package} (this may take a few minutes)...")
    try:
        install_engine(package)
        state.engine_installed = True
        _save_state(state)
        _log(f"  {engine_name} installed")
    except Exception as e:
        state.phase = PHASE_FAILED
        state.error = f"Failed to install {package}: {e}"
        state.finished_at = time.time()
        _save_state(state)
        _log(f"Setup failed: {state.error}")
        return state

    # Step 5: Download recommended model
    state.phase = PHASE_DOWNLOADING_MODEL
    model_key = _get_recommended_model()
    state.model_key = model_key
    _save_state(state)
    _log(f"Downloading model: {model_key}...")
    try:
        download_model(model_key)
        state.model_downloaded = True
        _save_state(state)
        _log(f"  Model {model_key} ready")
    except Exception as e:
        # Model download failure is non-fatal — engine still works,
        # user just gets a download on first serve.
        logger.warning("Model download failed (non-fatal): %s", e)
        _log(f"  Model download failed (will download on first use): {e}")

    # Done
    state.phase = PHASE_COMPLETE
    state.finished_at = time.time()
    _save_state(state)

    elapsed = state.finished_at - (state.started_at or state.finished_at)
    _log(f"Setup complete in {elapsed:.0f}s")
    return state



"""Shared helpers for CLI commands.

Extracted from cli.py to avoid circular imports and allow command modules
to share common utilities (auth, client creation, model completion, etc.).
"""

from __future__ import annotations

import logging
import os
import sys
import time
from typing import Any, Optional

import click
import httpx

_logger = logging.getLogger(__name__)

# Status codes safe to retry (transient server errors + rate limiting).
_RETRYABLE_STATUS_CODES = {502, 503, 504, 429}


def http_request(
    method: str,
    url: str,
    *,
    max_retries: int = 3,
    backoff_base: float = 0.5,
    timeout: float = 15.0,
    headers: Optional[dict[str, str]] = None,
    **kwargs: Any,
) -> httpx.Response:
    """HTTP request with automatic retry on transient failures.

    Retries on connection errors, timeouts, 502/503/504/429, and generic
    404s (FastAPI cold-start artifact).  Use this instead of raw
    ``httpx.get()`` / ``httpx.post()`` in CLI commands.
    """
    last_exc: Optional[Exception] = None
    for attempt in range(max_retries):
        try:
            with httpx.Client(timeout=timeout) as client:
                resp = client.request(method, url, headers=headers, **kwargs)

            if resp.status_code < 400:
                return resp

            # Generic 404 from cold start — retry.
            if (
                resp.status_code == 404
                and attempt < max_retries - 1
                and resp.text.strip() in ('{"detail":"Not Found"}', "Not Found")
            ):
                wait = backoff_base * (2**attempt)
                _logger.debug(
                    "Generic 404 on %s %s, retrying in %.1fs", method, url, wait
                )
                time.sleep(wait)
                continue

            # Retryable server error — backoff and retry.
            if (
                resp.status_code in _RETRYABLE_STATUS_CODES
                and attempt < max_retries - 1
            ):
                wait = backoff_base * (2**attempt)
                _logger.debug(
                    "HTTP %d on %s %s, retrying in %.1fs",
                    resp.status_code,
                    method,
                    url,
                    wait,
                )
                time.sleep(wait)
                continue

            return resp  # Non-retryable — let caller handle

        except (
            httpx.ConnectError,
            httpx.TimeoutException,
            httpx.RemoteProtocolError,
        ) as exc:
            last_exc = exc
            if attempt < max_retries - 1:
                wait = backoff_base * (2**attempt)
                click.echo(
                    f"  Connection failed, retrying ({attempt + 1}/{max_retries})...",
                    err=True,
                )
                time.sleep(wait)
            else:
                click.echo(
                    f"Error: failed to connect to Octomil after {max_retries} attempts: {exc}",
                    err=True,
                )
                sys.exit(1)

    # Should not reach here.
    click.echo(f"Error: request failed after {max_retries} attempts", err=True)
    sys.exit(1)


# ---------------------------------------------------------------------------
# Auth / credentials helpers
# ---------------------------------------------------------------------------


def _get_api_key() -> str:
    """Read API key from env or ~/.octomil/credentials (JSON or legacy format)."""
    import json

    key = os.environ.get("OCTOMIL_API_KEY", "")
    if not key:
        config_path = os.path.expanduser("~/.octomil/credentials")
        if os.path.exists(config_path):
            with open(config_path) as f:
                raw = f.read().strip()
            if not raw:
                return ""
            try:
                data = json.loads(raw)
                key = data.get("api_key", "")
            except (json.JSONDecodeError, ValueError):
                for line in raw.splitlines():
                    if line.startswith("api_key="):
                        key = line.split("=", 1)[1].strip()
                        break
    return key


def _require_api_key() -> str:
    key = _get_api_key()
    if not key:
        click.echo("No API key found. Run `octomil login` first.", err=True)
        sys.exit(1)
    return key


def _get_org_id() -> str | None:
    """Read org_id from env or ~/.octomil/credentials."""
    import json

    oid = os.environ.get("OCTOMIL_ORG_ID", "")
    if not oid:
        config_path = os.path.expanduser("~/.octomil/credentials")
        if os.path.exists(config_path):
            with open(config_path) as f:
                raw = f.read().strip()
            try:
                data = json.loads(raw)
                oid = data.get("org_id", "")
            except (json.JSONDecodeError, ValueError):
                pass
    return oid or None


def _get_client():  # type: ignore[no-untyped-def]
    from .client import OctomilClient

    return OctomilClient(api_key=_require_api_key(), org_id=_get_org_id())


def _get_telemetry_reporter():  # type: ignore[no-untyped-def]
    """Best-effort TelemetryReporter for funnel events. Returns None if no API key."""
    from .telemetry import TelemetryReporter

    api_key = _get_api_key()
    if not api_key:
        return None
    api_base: str = (
        os.environ.get("OCTOMIL_API_URL")
        or os.environ.get("OCTOMIL_API_BASE")
        or "https://api.octomil.com/api/v1"
    )
    return TelemetryReporter(api_key=api_key, api_base=api_base)


def _save_credentials(
    api_key: str,
    org: Optional[str] = None,
    org_id: Optional[str] = None,
) -> None:
    """Save credentials to ~/.octomil/credentials as JSON with restrictive permissions."""
    import json

    config_dir = os.path.expanduser("~/.octomil")
    os.makedirs(config_dir, exist_ok=True)
    config_path = os.path.join(config_dir, "credentials")
    data: dict[str, str] = {"api_key": api_key}
    if org:
        data["org_name"] = org
    if org_id:
        data["org_id"] = org_id
    with open(config_path, "w") as f:
        json.dump(data, f, indent=2)
        f.write("\n")
    os.chmod(config_path, 0o600)


# ---------------------------------------------------------------------------
# Enterprise helpers
# ---------------------------------------------------------------------------


def _get_enterprise_client(api_base: Optional[str] = None):  # type: ignore[no-untyped-def]
    """Build an EnterpriseClient with the current API key."""
    from .enterprise import EnterpriseClient

    key = _require_api_key()
    return EnterpriseClient(api_key=key, api_base=api_base)


def _require_org_id() -> str:
    """Read org_id from config/env or exit."""
    from .enterprise import get_org_id

    org_id = get_org_id()
    if not org_id:
        click.echo(
            "No org_id configured. Run `octomil init <name>` first, or "
            "set OCTOMIL_ORG_ID.",
            err=True,
        )
        sys.exit(1)
    return org_id


# ---------------------------------------------------------------------------
# Auto-optimization helpers
# ---------------------------------------------------------------------------

# Quant variants that can appear after ":" in a model tag
_KNOWN_QUANT_SUFFIXES = frozenset(
    {
        "q2_k",
        "q3_k_s",
        "q3_k_m",
        "q4_0",
        "q4_k_s",
        "q4_k_m",
        "q5_0",
        "q5_k_s",
        "q5_k_m",
        "q6_k",
        "q8_0",
        "4bit",
        "8bit",
        "fp16",
        "f16",
        "f32",
    }
)


# ~200 most common model names for offline shell autocomplete.
# Covers 99%+ of what users type. The API fallback below only adds
# user-specific custom model IDs from their registry.
_KNOWN_MODEL_NAMES = frozenset(
    {
        # Llama (Meta)
        "llama2",
        "llama2-7b",
        "llama2-13b",
        "llama2-70b",
        "llama2-uncensored",
        "llama3",
        "llama3-8b",
        "llama3-70b",
        "llama3.1",
        "llama3.1-8b",
        "llama3.1-70b",
        "llama3.1-405b",
        "llama3.2",
        "llama3.2-1b",
        "llama3.2-3b",
        "llama3.2-vision",
        "llama3.3",
        "llama3.3-70b",
        "llama4",
        "llama4-scout",
        "llama4-maverick",
        "llama-guard3",
        # Octomil catalog aliases (dash style)
        "llama-1b",
        "llama-3b",
        "llama-8b",
        "llama-3.2-1b",
        "llama-3.2-3b",
        "llama-3.1-8b",
        # CodeLlama
        "codellama",
        "codellama-7b",
        "codellama-13b",
        "codellama-34b",
        "codellama-python",
        "codellama-instruct",
        # Gemma (Google)
        "gemma",
        "gemma-2b",
        "gemma-7b",
        "gemma2",
        "gemma2-2b",
        "gemma2-9b",
        "gemma2-27b",
        "gemma3",
        "gemma3-1b",
        "gemma3-4b",
        "gemma3-12b",
        "gemma3-27b",
        "gemma3n",
        "gemma3n-e2b",
        "gemma3n-e4b",
        "codegemma",
        # Octomil catalog keys/aliases
        "gemma-1b",
        "gemma-4b",
        "gemma-12b",
        "gemma-27b",
        "gemma-3-1b",
        "gemma-3-4b",
        "gemma-3-12b",
        "gemma-3-27b",
        "gemma-3b",
        # Phi (Microsoft)
        "phi",
        "phi-2",
        "phi3",
        "phi3-mini",
        "phi3.5",
        "phi4",
        "phi4-mini",
        "phi4-reasoning",
        "phi4-mini-reasoning",
        "phi-4",
        "phi-mini",
        "phi-4-mini",
        "phi-3.5-mini",
        "phi-mini-3.8b",
        # Qwen (Alibaba)
        "qwen",
        "qwen2",
        "qwen2.5",
        "qwen2.5-0.5b",
        "qwen2.5-1.5b",
        "qwen2.5-3b",
        "qwen2.5-7b",
        "qwen2.5-14b",
        "qwen2.5-32b",
        "qwen2.5-72b",
        "qwen2.5-coder",
        "qwen3",
        "qwen3-0.6b",
        "qwen3-1.7b",
        "qwen3-4b",
        "qwen3-8b",
        "qwen3-14b",
        "qwen3-32b",
        "qwen3-coder",
        "qwen3.5",
        "qwq",
        "codeqwen",
        # Octomil catalog keys/aliases
        "qwen-1.5b",
        "qwen-3b",
        "qwen-7b",
        "qwen-moe-14b",
        "qwen-2.5-1.5b",
        "qwen-2.5-3b",
        "qwen-2.5-7b",
        "qwen-1.5-moe",
        "qwen-moe",
        "qwen3-1b",
        # Mistral / Mixtral
        "mistral",
        "mistral-7b",
        "mistral-nemo",
        "mistral-small",
        "mistral-small3.1",
        "mistral-large",
        "mixtral",
        "mixtral-8x7b",
        "mixtral-8x22b",
        "mixtral-instruct",
        "mixtral-8x22b-instruct",
        "codestral",
        "devstral",
        "magistral",
        "mathstral",
        "ministral-3",
        # DeepSeek
        "deepseek-r1",
        "deepseek-r1-1.5b",
        "deepseek-r1-7b",
        "deepseek-r1-8b",
        "deepseek-r1-14b",
        "deepseek-r1-32b",
        "deepseek-r1-70b",
        "deepseek-v2",
        "deepseek-v2-lite",
        "deepseek-v2.5",
        "deepseek-v3",
        "deepseek-v3.1",
        "deepseek-coder",
        "deepseek-coder-v2",
        "deepseek-v2-lite-chat",
        # Whisper (OpenAI)
        "whisper-tiny",
        "whisper-base",
        "whisper-small",
        "whisper-medium",
        "whisper-large",
        "whisper-large-v2",
        "whisper-large-v3",
        "whisper-turbo",
        # Nous / Hermes
        "nous-hermes",
        "nous-hermes2",
        "hermes3",
        "openhermes",
        # Dolphin
        "dolphin3",
        "dolphin-llama3",
        "dolphin-mistral",
        "dolphin-phi",
        # Yi (01.AI)
        "yi",
        "yi-coder",
        "yi-34b",
        # Falcon (TII)
        "falcon",
        "falcon2",
        "falcon3",
        # Command-R (Cohere)
        "command-r",
        "command-r-plus",
        "command-r7b",
        # StarCoder
        "starcoder",
        "starcoder2",
        "starcoder2-3b",
        "starcoder2-15b",
        # SmolLM (HuggingFace)
        "smollm",
        "smollm2",
        "smollm-360m",
        "smollm-1.7b",
        "smollm2-360m",
        "smollm2-1.7b",
        # TinyLlama
        "tinyllama",
        # Vicuna / WizardLM
        "vicuna",
        "wizardlm2",
        "wizardcoder",
        # GLM (Zhipu)
        "glm4",
        "glm-4.7",
        # Granite (IBM)
        "granite-code",
        "granite3.2",
        "granite4",
        # Aya (Cohere)
        "aya-expanse",
        # OLMo (AI2)
        "olmo2",
        # Multimodal / Vision
        "llava",
        "llava-llama3",
        "moondream",
        "minicpm-v",
        # Embedding
        "nomic-embed-text",
        "mxbai-embed-large",
        "bge-m3",
        "all-minilm",
        "snowflake-arctic-embed",
        # Stable / Flux
        "stable-code",
        "stablelm2",
        # Other notable
        "dbrx",
        "dbrx-instruct",
        "nemotron",
        "nemotron-mini",
        "neural-chat",
        "openchat",
        "solar",
        "solar-pro",
        "sqlcoder",
        "zephyr",
        "orca-mini",
        "orca2",
        "tinydolphin",
        "internlm2",
    }
)


def _complete_model_name(ctx, param, incomplete):
    """Shell completion callback for model name arguments/options."""
    from click.shell_completion import CompletionItem

    names = set(_KNOWN_MODEL_NAMES)

    # Best-effort: include user's custom models from their registry.
    try:
        import httpx

        api_key = _get_api_key()
        api_base = (
            os.environ.get("OCTOMIL_API_URL")
            or os.environ.get("OCTOMIL_API_BASE")
            or "https://api.octomil.com/api/v1"
        )
        if api_key:
            resp = httpx.get(
                f"{api_base}/models",
                headers={"Authorization": f"Bearer {api_key}"},
                timeout=2.0,
            )
            if resp.status_code == 200:
                for m in resp.json().get("models", []):
                    mid = m.get("model_id") or m.get("id", "")
                    if mid:
                        names.add(mid)
    except Exception:
        pass

    return [CompletionItem(n) for n in sorted(names) if n.startswith(incomplete)]


def _has_explicit_quant(model_tag: str) -> bool:
    """Check if the model tag already specifies a quantization variant."""
    if ":" not in model_tag:
        return False
    variant = model_tag.rsplit(":", 1)[1].lower().replace("-", "_")
    return variant in _KNOWN_QUANT_SUFFIXES


def _auto_optimize(model_tag: str, context_length: int = 4096) -> str | None:
    """Run hardware-aware optimization and print results.

    Returns the recommended quantization string (e.g. "Q6_K"), or ``None``
    if the model size cannot be determined.  Does NOT modify *model_tag*.
    """
    from .hardware import detect_hardware
    from .model_optimizer import ModelOptimizer, _resolve_model_size

    model_size_b = _resolve_model_size(model_tag)
    if model_size_b is None:
        return None

    hw = detect_hardware()
    opt = ModelOptimizer(hw)
    config = opt.pick_quant_and_offload(model_size_b, context_length)
    speed = opt.predict_speed(model_size_b, config)

    click.echo()
    click.secho("  Hardware optimization", bold=True)
    click.echo(f"    Quantization: {config.quantization}")
    click.echo(
        f"    Strategy: {config.strategy.value} ({config.gpu_layers} GPU layers)"
    )
    click.echo(f"    VRAM: {config.vram_gb:.1f} GB  RAM: {config.ram_gb:.1f} GB")
    click.echo(
        f"    Est. speed: {speed.tokens_per_second:.1f} tok/s"
        f" ({speed.confidence} confidence)"
    )
    if config.warning:
        click.secho(f"    Warning: {config.warning}", fg="yellow")

    return config.quantization


# ---------------------------------------------------------------------------
# Welcome message
# ---------------------------------------------------------------------------

WELCOME_MESSAGE = """\
Octomil — on-device AI inference

  Get started:
    1. octomil serve phi-4-mini              run a model locally
    2. octomil benchmark phi-4-mini          measure performance
    3. octomil deploy phi-4-mini --phone     ship to devices
    4. octomil login                         unlock fleet features

  Useful commands:
    octomil models                           list available models
    octomil push <file> --version 1.0.0      upload a model
    octomil dashboard                        open web dashboard
    octomil launch <agent>                   launch coding agent

  Run octomil <command> --help for details.
  Docs: https://docs.octomil.com
"""



"""Batch settlement for x402 micro-payments.

Accumulates signed EIP-3009 TransferWithAuthorization payloads in-memory
and submits them as a single batch when the running total hits a configurable
threshold (default: $1 USDC = 1,000,000 base units).  One multicall
transaction amortises gas across ~1000 micro-payments.
"""

from __future__ import annotations

import logging
import threading
import time
from collections import OrderedDict
from dataclasses import dataclass, field
from typing import Any

logger = logging.getLogger(__name__)


@dataclass
class PendingAuthorization:
    """A verified but unsettled payment authorization."""

    authorization: dict[str, Any]
    signature: str
    payer: str
    amount: int  # base units (e.g. 1000 = 0.001 USDC)
    request_path: str
    timestamp: float = field(default_factory=time.time)


class SettlementStore:
    """Thread-safe in-memory store for pending payment authorizations.

    Mirrors the ``NonceTracker`` pattern — all access guarded by a lock.
    """

    def __init__(self, threshold: int = 1_000_000) -> None:
        self._pending: OrderedDict[str, PendingAuthorization] = OrderedDict()
        self._lock = threading.Lock()
        self._total: int = 0
        self._threshold: int = threshold
        self._settled_count: int = 0
        self._settled_total: int = 0

    def add(self, nonce: str, auth: PendingAuthorization) -> bool:
        """Store a pending authorization.

        Returns True if the running total has reached the settlement threshold.
        """
        with self._lock:
            self._pending[nonce] = auth
            self._total += auth.amount
            return self._total >= self._threshold

    def pop_batch(self) -> list[PendingAuthorization]:
        """Atomically drain all pending authorizations.

        Resets the running total and increments lifetime counters.
        """
        with self._lock:
            batch = list(self._pending.values())
            self._settled_count += len(batch)
            self._settled_total += self._total
            self._pending.clear()
            self._total = 0
            return batch

    def requeue(self, items: list[PendingAuthorization]) -> None:
        """Re-add authorizations after a failed settlement attempt."""
        with self._lock:
            for item in items:
                nonce = item.authorization.get("nonce", "")
                if isinstance(nonce, bytes):
                    nonce = nonce.hex()
                nonce = str(nonce)
                if nonce not in self._pending:
                    self._pending[nonce] = item
                    self._total += item.amount

    def discard(self, nonce: str) -> None:
        """Remove a single authorization (e.g. on service error / refund)."""
        with self._lock:
            auth = self._pending.pop(nonce, None)
            if auth is not None:
                self._total -= auth.amount

    def stats(self) -> dict[str, Any]:
        """Return settlement statistics."""
        with self._lock:
            return {
                "pending_count": len(self._pending),
                "pending_total": self._total,
                "settled_count": self._settled_count,
                "settled_total": self._settled_total,
                "threshold": self._threshold,
            }


async def settle_batch(
    store: SettlementStore,
    facilitator_url: str,
    config: Any,
) -> None:
    """Submit a batch of authorizations for on-chain settlement.

    If ``facilitator_url`` is set, POSTs the batch to ``{facilitator_url}/settle``.
    Otherwise logs the batch for manual settlement (the signed authorizations
    remain valid on-chain until their ``validBefore`` timestamp).

    On facilitator failure the entire batch is re-queued into the store.
    """
    batch = store.pop_batch()
    if not batch:
        return

    logger.info(
        "x402: settling batch of %d authorizations (total=%d base units)",
        len(batch),
        sum(a.amount for a in batch),
    )

    if not facilitator_url:
        logger.info(
            "x402: no facilitator_url configured — %d authorizations logged for manual settlement",
            len(batch),
        )
        return

    payload = {
        "network": getattr(config, "network", "base"),
        "chainId": getattr(config, "resolved_chain_id", lambda: 8453)(),
        "tokenContract": getattr(config, "resolved_token_contract", lambda: "")(),
        "authorizations": [
            {
                "authorization": a.authorization,
                "signature": a.signature,
                "payer": a.payer,
                "amount": a.amount,
            }
            for a in batch
        ],
    }

    try:
        import httpx

        async with httpx.AsyncClient(timeout=30.0) as client:
            resp = await client.post(f"{facilitator_url}/settle", json=payload)
            resp.raise_for_status()
            logger.info("x402: facilitator accepted batch (%d auths)", len(batch))
    except Exception as exc:
        logger.error("x402: facilitator settlement failed: %s — re-queuing %d auths", exc, len(batch))
        store.requeue(batch)



"""Serve command — local OpenAI-compatible inference server."""

from __future__ import annotations

import os
import sys

import click

from octomil.cli_helpers import (
    _auto_optimize,
    _complete_model_name,
    _get_api_key,
    _get_telemetry_reporter,
    _has_explicit_quant,
)


def register(cli: click.Group) -> None:
    cli.add_command(serve)


@click.command()
@click.argument("model", shell_complete=_complete_model_name)
@click.option("--port", "-p", default=8080, help="Port to listen on.")
@click.option("--host", default="0.0.0.0", help="Host to bind to.")
@click.option("--benchmark", is_flag=True, help="Run latency benchmark on startup.")
@click.option("--share", is_flag=True, help="Share anonymous benchmark data with Octomil.")
@click.option(
    "--json-mode",
    is_flag=True,
    help="Default all responses to JSON output (response_format=json_object).",
)
@click.option(
    "--cache-size",
    default=2048,
    type=int,
    help="KV cache size in MB (default: 2048). Caches prompt prefixes to speed up multi-turn conversations.",
)
@click.option(
    "--no-cache",
    is_flag=True,
    help="Disable KV cache entirely.",
)
@click.option(
    "--engine",
    "-e",
    default=None,
    help="Force a specific engine (mlx-lm, llama.cpp, mnn, onnxruntime, ollama). "
    "Default: auto-benchmark all available engines and pick fastest.",
)
@click.option(
    "--models",
    default=None,
    help="Comma-separated list of models to load (ordered small to large). "
    "Enables multi-model serving. Example: smollm-360m,phi-mini,llama-3b",
)
@click.option(
    "--auto-route",
    is_flag=True,
    help="Enable automatic query routing across loaded models. " "Requires --models with 2+ models.",
)
@click.option(
    "--route-strategy",
    default="complexity",
    type=click.Choice(["complexity"]),
    help="Routing strategy for --auto-route (default: complexity).",
)
@click.option(
    "--max-queue",
    default=32,
    type=int,
    help="Max pending requests in the queue (default: 32). Set to 0 to disable.",
)
@click.option(
    "--compress-context",
    is_flag=True,
    help="Enable prompt compression. Compresses long prompts before inference "
    "to reduce context window usage and speed up prefill.",
)
@click.option(
    "--compression-strategy",
    default="token_pruning",
    type=click.Choice(["token_pruning", "sliding_window"]),
    help="Compression strategy (default: token_pruning). "
    "token_pruning removes low-information tokens. "
    "sliding_window keeps recent turns verbatim and summarises older ones.",
)
@click.option(
    "--compression-ratio",
    default=0.5,
    type=float,
    help="Target compression ratio for token pruning (0.0-1.0, default: 0.5). "
    "Higher values prune more aggressively.",
)
@click.option(
    "--compression-max-turns",
    default=4,
    type=int,
    help="Number of recent conversation turns to keep verbatim " "when using sliding_window strategy (default: 4).",
)
@click.option(
    "--compression-threshold",
    default=256,
    type=int,
    help="Minimum estimated token count before compression kicks in (default: 256).",
)
@click.option(
    "--tool-use",
    is_flag=True,
    default=False,
    help="Pre-load coding agent tool schemas for structured output. "
    "Exposes tool definitions (read_file, write_file, edit_file, run_command, "
    "search_files) at /v1/tool-schemas for coding agents like Aider, Goose, "
    "and OpenCode.",
)
@click.option(
    "--early-exit-threshold",
    default=None,
    type=float,
    help="Enable early exit with this entropy threshold (0.0-1.0). "
    "Tokens exit early when intermediate logit entropy drops below this value. "
    "Lower = fewer exits (conservative), higher = more exits (aggressive). "
    "Example: --early-exit-threshold 0.3",
)
@click.option(
    "--speed-quality",
    default=None,
    type=click.Choice(["quality", "balanced", "fast"]),
    help="Speed-quality preset for early exit. "
    "quality: conservative (threshold=0.1), "
    "balanced: moderate (threshold=0.3), "
    "fast: aggressive (threshold=0.5). "
    "Overridden by --early-exit-threshold if both are set.",
)
def serve(
    model: str,
    port: int,
    host: str,
    benchmark: bool,
    share: bool,
    json_mode: bool,
    cache_size: int,
    no_cache: bool,
    engine: str | None,
    models: str | None,
    auto_route: bool,
    route_strategy: str,
    max_queue: int,
    compress_context: bool,
    compression_strategy: str,
    compression_ratio: float,
    compression_max_turns: int,
    compression_threshold: int,
    tool_use: bool,
    early_exit_threshold: float | None,
    speed_quality: str | None,
) -> None:
    """Start a local OpenAI-compatible inference server.

    MODEL accepts Ollama-style model:variant syntax:

    \b
        octomil serve gemma-1b              # auto-picks best quant for your hw
        octomil serve gemma-1b:8bit         # explicit 8-bit quantization
        octomil serve llama-8b:fp16         # full precision (no auto-optimize)
        octomil serve llama-3b:q4_k_m       # engine-specific quant (explicit)

    Without an explicit quantization variant, Octomil detects your GPU/RAM
    and picks the best quantization automatically.

    Auto-detects all available inference engines, benchmarks each,
    and picks the fastest for your hardware. Override with --engine.

    \b
    Multi-model with auto-routing:
        octomil serve smollm-360m --models smollm-360m,phi-mini,llama-3b --auto-route

    Simple queries (greetings, short factual) route to the smallest model.
    Complex queries (code, reasoning, multi-step) route to the largest.
    If a model fails, the next larger model is tried automatically.

    \b
    Force a specific engine:
        octomil serve gemma-1b --engine llama.cpp

    \b
    Use --json-mode to default all responses to valid JSON output:
        octomil serve gemma-1b --json-mode
    """
    api_key = _get_api_key() if share else None
    api_base: str = (
        os.environ.get("OCTOMIL_API_URL") or os.environ.get("OCTOMIL_API_BASE") or "https://api.octomil.com/api/v1"
    )
    cache_enabled = not no_cache

    # Determine if multi-model mode
    model_list: list[str] | None = None
    if models:
        model_list = [m.strip() for m in models.split(",") if m.strip()]

    if auto_route and not model_list:
        click.echo(
            "Error: --auto-route requires --models with 2+ models.",
            err=True,
        )
        sys.exit(1)

    if auto_route and model_list and len(model_list) < 2:
        click.echo(
            "Error: --auto-route requires at least 2 models in --models.",
            err=True,
        )
        sys.exit(1)

    # Multi-model routing mode
    if auto_route and model_list:
        _serve_multi_model(
            model_list=model_list,
            port=port,
            host=host,
            api_key=api_key,
            api_base=api_base,
            json_mode=json_mode,
            cache_size=cache_size,
            cache_enabled=cache_enabled,
            engine=engine,
            route_strategy=route_strategy,
        )
        return

    # Check if this is a whisper (speech-to-text) model
    from octomil.engines.whisper_engine import is_whisper_model

    is_whisper = is_whisper_model(model)

    # Auto-optimize: pick best quantization for hardware if not explicit
    if not is_whisper and not _has_explicit_quant(model):
        best_quant = _auto_optimize(model, context_length=cache_size)
        if best_quant:
            model = f"{model}:{best_quant.lower()}"
            click.echo(f"    Serving as: {model}")

    # Single-model mode (original behaviour)
    _print_engine_detection(model, engine)

    click.echo(f"\nStarting Octomil serve on {host}:{port}")
    if is_whisper:
        click.echo(f"Model: {model} (speech-to-text)")
        click.echo(f"POST http://localhost:{port}/v1/audio/transcriptions")
    else:
        click.echo(f"Model: {model}")
        if engine:
            click.echo(f"Engine: {engine} (manual override)")
        if json_mode:
            click.echo("JSON mode: enabled (all responses default to valid JSON)")
        click.echo(f"OpenAI-compatible API: http://localhost:{port}/v1/chat/completions")
    click.echo(f"Engine info: http://localhost:{port}/v1/engines")
    click.echo(f"Health check: http://localhost:{port}/health")
    if not is_whisper:
        if cache_enabled:
            click.echo(f"KV cache: enabled ({cache_size} MB)")
            click.echo(f"Cache stats: http://localhost:{port}/v1/cache/stats")
        else:
            click.echo("KV cache: disabled")
        if max_queue > 0:
            click.echo(f"Request queue: enabled (max_depth={max_queue})")
            click.echo(f"Queue stats: http://localhost:{port}/v1/queue/stats")
        else:
            click.echo("Request queue: disabled")
        if compress_context:
            click.echo(
                f"Context compression: enabled "
                f"(strategy={compression_strategy}, "
                f"ratio={compression_ratio}, "
                f"threshold={compression_threshold} tokens)"
            )
        if tool_use:
            click.echo("Tool-use mode: enabled (coding agent tool schemas loaded)")
            click.echo(f"Tool schemas: http://localhost:{port}/v1/tool-schemas")

    # Build early exit config
    from octomil.early_exit import config_from_cli as _ee_config_from_cli

    ee_config = _ee_config_from_cli(
        early_exit_threshold=early_exit_threshold,
        speed_quality=speed_quality,
    )

    if not is_whisper and ee_config.enabled:
        preset_label = f" (preset: {ee_config.preset.value})" if ee_config.preset else ""
        click.echo(
            f"Early exit: enabled (threshold={ee_config.effective_threshold:.2f}"
            f", min_layers_frac={ee_config.effective_min_layers_fraction:.2f}"
            f"{preset_label})"
        )
        click.echo(f"Early exit stats: http://localhost:{port}/v1/early-exit/stats")

    if benchmark:
        click.echo("Benchmark mode: will run latency test after model loads.")

    if share and not api_key:
        click.echo(
            "Warning: --share requires an API key to upload benchmark data. "
            "Run `octomil login` or set OCTOMIL_API_KEY.",
            err=True,
        )

    from octomil.serve import run_server

    run_server(
        model,
        port=port,
        host=host,
        api_key=api_key,
        api_base=api_base,
        json_mode=json_mode,
        cache_size_mb=cache_size,
        cache_enabled=cache_enabled,
        engine=engine,
        max_queue_depth=max_queue,
        compress_context=compress_context,
        compression_strategy=compression_strategy,
        compression_ratio=compression_ratio,
        compression_max_turns=compression_max_turns,
        compression_threshold=compression_threshold,
        tool_use=tool_use,
        early_exit_config=ee_config if ee_config.enabled else None,
    )


def _serve_multi_model(
    model_list: list[str],
    port: int,
    host: str,
    api_key: str | None,
    api_base: str,
    json_mode: bool,
    cache_size: int,
    cache_enabled: bool,
    engine: str | None,
    route_strategy: str,
) -> None:
    """Start multi-model serving with query routing."""
    from octomil.routing import assign_tiers

    tiers = assign_tiers(model_list)
    tier_labels = {
        "fast": "<0.3 complexity",
        "balanced": "0.3-0.7 complexity",
        "quality": ">0.7 complexity",
    }

    click.echo(f"\nLoading {len(model_list)} models for auto-routing...")
    for name in model_list:
        info = tiers[name]
        label = tier_labels.get(info.tier, info.tier)
        click.echo(f"  {name}: tier={info.tier} ({label})")

    click.echo(f"\nRouting enabled (strategy: {route_strategy})")
    click.echo(f"Serving on {host}:{port}")
    click.echo(f"OpenAI-compatible API: http://localhost:{port}/v1/chat/completions")
    click.echo(f"Routing stats: http://localhost:{port}/v1/routing/stats")
    click.echo(f"Health check: http://localhost:{port}/health")
    if json_mode:
        click.echo("JSON mode: enabled (all responses default to valid JSON)")

    from octomil.serve import run_multi_model_server

    run_multi_model_server(
        model_list,
        port=port,
        host=host,
        api_key=api_key,
        api_base=api_base,
        json_mode=json_mode,
        cache_size_mb=cache_size,
        cache_enabled=cache_enabled,
        engine=engine,
        route_strategy=route_strategy,
    )


def _prompt_engine_install() -> bool:
    """Prompt user to install an inference engine. Recommends max performance first.

    Platform-specific recommendations:
    - macOS Apple Silicon: MLX (fastest) → llama.cpp → Ollama fallback
    - macOS Intel: llama.cpp → ONNX Runtime → Ollama fallback
    - Linux: llama.cpp → ONNX Runtime → Ollama fallback
    - Windows: llama.cpp → Ollama fallback

    Returns True if an engine was installed and is ready.
    """
    import platform as _platform
    import shutil
    import subprocess

    system = _platform.system()
    machine = _platform.machine()
    is_apple_silicon = system == "Darwin" and machine == "arm64"

    click.echo(
        click.style(
            "\nNo inference engines found.",
            fg="yellow",
        )
    )

    # Check if Ollama binary is on PATH but not running
    ollama_installed = shutil.which("ollama") is not None

    # Build platform-specific recommendation list (best performance first)
    recommendations: list[tuple[str, str]] = []  # (label, install_cmd)
    if is_apple_silicon:
        recommendations.append(("MLX — fastest on Apple Silicon", "pip install octomil-sdk[mlx]"))
        recommendations.append(("llama.cpp — good alternative", "pip install llama-cpp-python"))
    elif system == "Darwin":
        recommendations.append(("llama.cpp", "pip install llama-cpp-python"))
        recommendations.append(("ONNX Runtime", "pip install onnxruntime"))
    elif system == "Linux":
        recommendations.append(("llama.cpp", "pip install llama-cpp-python"))
        recommendations.append(("ONNX Runtime", "pip install onnxruntime"))
    else:
        recommendations.append(("llama.cpp", "pip install llama-cpp-python"))

    # Ollama always last — zero-pip fallback
    if system == "Darwin":
        recommendations.append(("Ollama — no pip needed", "brew install ollama"))
    elif system == "Linux":
        recommendations.append(("Ollama — no pip needed", "curl -fsSL https://ollama.com/install.sh | sh"))
    else:
        recommendations.append(("Ollama — no pip needed", "Visit https://ollama.com/download"))

    if not sys.stdin.isatty():
        # Non-interactive: print all options
        if ollama_installed:
            click.echo("  Ollama is installed but not running.\n")
            click.echo("    ollama serve\n")
        click.echo("  Install an inference engine:\n")
        for label, cmd in recommendations:
            click.echo(f"    {cmd:<45s} # {label}")
        click.echo()
        return False

    # Interactive: if Ollama is installed but not running, offer to start it
    if ollama_installed:
        click.echo("  Ollama is installed but not running.\n")
        if click.confirm("  Start Ollama now?", default=True):
            try:
                subprocess.Popen(
                    ["ollama", "serve"],
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                )
                import time

                time.sleep(2)  # Give Ollama a moment to start
                from octomil.ollama import is_ollama_running

                if is_ollama_running():
                    click.echo(click.style("  Ollama started.\n", fg="green"))
                    return True
            except Exception:
                pass
            click.echo(
                click.style("  Could not start Ollama.", fg="red"),
                err=True,
            )
            click.echo("  Start manually: ollama serve\n")
        return False

    # Interactive: show recommendations (best performance first)
    top_label, top_cmd = recommendations[0]
    click.echo(f"  Recommended: {top_label}\n")

    if click.confirm(f"  Install now? ({top_cmd})", default=True):
        click.echo()
        # Parse pip command
        if top_cmd.startswith("pip install"):
            pkg = top_cmd.split("pip install ", 1)[1]
            if getattr(sys, "frozen", False):
                pip_cmd = ["pip", "install", pkg]
            else:
                pip_cmd = [sys.executable, "-m", "pip", "install", pkg]
            try:
                subprocess.check_call(pip_cmd)
                click.echo(click.style(f"\n  {pkg} installed successfully.\n", fg="green"))
                try:
                    reporter = _get_telemetry_reporter()
                    if reporter:
                        reporter.report_funnel_event("cli_install", success=True)
                except Exception:
                    pass
                return True
            except subprocess.CalledProcessError:
                click.echo(
                    click.style(f"\n  Failed to install {pkg}.", fg="red"),
                    err=True,
                )
        else:
            click.echo(f"  Run: {top_cmd}\n")
    else:
        click.echo("\n  Other options:\n")
        for label, cmd in recommendations:
            click.echo(f"    {cmd:<45s} # {label}")
        click.echo()

    return False


def _print_engine_detection(model: str, engine_override: str | None) -> None:
    """Print engine detection results to terminal."""
    from octomil.engines import get_registry

    registry = get_registry()

    click.echo("\nDetecting engines...")
    detections = registry.detect_all(model)
    for d in detections:
        if d.engine.name == "echo":
            continue
        if d.available:
            info = f" ({d.info})" if d.info else ""
            click.echo(click.style(f"  + {d.engine.display_name}{info}", fg="green"))
        else:
            click.echo(click.style(f"  - {d.engine.display_name}", dim=True))

    if engine_override:
        click.echo(f"\nUsing {engine_override} (manual override)")
        return

    available = [d for d in detections if d.available and d.engine.name != "echo"]
    if len(available) > 1:
        click.echo(f"\nBenchmarking {model} across {len(available)} engines...")
        click.echo("(this runs a quick 32-token generation on each)")
    elif len(available) == 1:
        click.echo(f"\nUsing {available[0].engine.display_name} (only available engine)")
    else:
        # Frozen binary: try re-exec into managed venv before prompting
        if getattr(sys, "frozen", False):
            reexeced = _try_venv_reexec()
            if reexeced:
                return  # os.execv replaces process; this line is unreachable

        installed = _prompt_engine_install()
        if installed:
            detections = registry.detect_all(model)
            available = [d for d in detections if d.available and d.engine.name != "echo"]
            if available:
                click.echo(f"  Using {available[0].engine.display_name}")
                return
        click.echo(
            click.style(
                "  Using echo backend for now (mirrors input, no real inference).\n",
                dim=True,
            )
        )


def _try_venv_reexec() -> bool:
    """Try to re-exec into the managed venv's Python for native engine support.

    When running from a frozen binary with no native engines, this checks
    if ``octomil setup`` has prepared a venv with mlx-lm or llama.cpp.
    If so, ``os.execv()`` replaces this process entirely with the venv's
    Python running ``octomil serve``. Single process, no proxy.

    If no venv exists and setup hasn't run yet, runs setup inline so that
    ``octomil serve`` is a single-command experience with no separate step.

    Returns True if re-exec was initiated (unreachable after os.execv).
    Returns False if setup failed and we should fall through.
    """
    from octomil.setup import (
        get_venv_python,
        is_engine_ready,
        is_setup_in_progress,
        load_state,
        run_setup,
    )

    if is_setup_in_progress():
        _wait_for_setup()

    venv_py = get_venv_python()
    if not venv_py or not is_engine_ready():
        # No venv yet — run setup inline so serve just works
        state = load_state()
        if state.phase != "failed":
            click.echo(
                click.style(
                    "\n  First run: setting up native inference engine...\n",
                    fg="cyan",
                )
            )
            result = run_setup()
            if result.phase == "failed":
                click.echo(click.style(f"  Setup failed: {result.error}", fg="red"))
                return False
            # Re-check after setup
            venv_py = get_venv_python()
            if not venv_py or not is_engine_ready():
                return False
        else:
            return False

    # Build the argv for re-exec: venv python -m octomil serve <original args>
    # Reconstruct original args from sys.argv (frozen binary: ["octomil", "serve", ...])
    # We need everything after "serve" in the original invocation.
    original_args = sys.argv[1:]  # ["serve", "model", "--port", "8080", ...]
    new_argv = [venv_py, "-m", "octomil", *original_args]

    click.echo(
        click.style(
            "\n  Re-launching with native engine via managed venv...\n",
            fg="green",
        )
    )
    os.execv(venv_py, new_argv)
    return True  # unreachable after execv


def _wait_for_setup() -> None:
    """Wait for a running ``octomil setup`` to finish, showing progress."""
    import time

    from octomil.setup import is_setup_in_progress, load_state

    click.echo("\n  Engine setup is in progress, waiting...")
    phase_labels = {
        "creating_venv": "creating virtual environment",
        "installing_engine": "installing inference engine",
        "downloading_model": "downloading model",
    }

    last_phase = ""
    waited = 0
    while is_setup_in_progress() and waited < 600:
        state = load_state()
        label = phase_labels.get(state.phase, state.phase)
        if state.phase != last_phase:
            click.echo(f"    {label}...")
            last_phase = state.phase
        time.sleep(2)
        waited += 2

    state = load_state()
    if state.phase == "complete":
        click.echo(click.style("  Setup complete.", fg="green"))
    elif state.phase == "failed":
        click.echo(click.style(f"  Setup failed: {state.error}", fg="red"))



"""Tests for the ``octomil federation`` CLI command group."""

from __future__ import annotations

from unittest.mock import MagicMock, patch

from click.testing import CliRunner

from octomil.cli import main


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


def _mock_client(api_responses: dict | None = None):
    """Build a mock OctomilClient with a mock _api that returns preconfigured data.

    ``api_responses`` maps ``(method, path_prefix)`` tuples to return values.
    For simple cases you can also just set return values on the mock directly.
    """
    mock = MagicMock()
    mock._org_id = "org_test_123"

    api = MagicMock()
    if api_responses:

        def _get(path, params=None):
            for (m, prefix), val in api_responses.items():
                if m == "get" and path.startswith(prefix):
                    return val
            return []

        def _post(path, payload=None):
            for (m, prefix), val in api_responses.items():
                if m == "post" and path.startswith(prefix):
                    return val
            return {}

        api.get = MagicMock(side_effect=_get)
        api.post = MagicMock(side_effect=_post)

    mock._api = api
    return mock


# ---------------------------------------------------------------------------
# federation create
# ---------------------------------------------------------------------------


class TestFederationCreate:
    @patch("octomil.commands.federation._get_client")
    def test_create_success(self, mock_get_client, monkeypatch):
        monkeypatch.setenv("OCTOMIL_API_KEY", "test-key")
        client = _mock_client()
        client._api.post.return_value = {"id": "fed_abc123", "name": "my-fed"}
        mock_get_client.return_value = client

        runner = CliRunner()
        result = runner.invoke(
            main,
            ["federation", "create", "my-fed", "--description", "Test federation"],
        )

        assert result.exit_code == 0
        assert "Federation created: my-fed" in result.output
        assert "fed_abc123" in result.output

        # Verify the API was called with the right payload
        client._api.post.assert_called_once_with(
            "/federations",
            {
                "name": "my-fed",
                "org_id": "org_test_123",
                "description": "Test federation",
            },
        )

    @patch("octomil.commands.federation._get_client")
    def test_create_without_description(self, mock_get_client, monkeypatch):
        monkeypatch.setenv("OCTOMIL_API_KEY", "test-key")
        client = _mock_client()
        client._api.post.return_value = {"id": "fed_xyz"}
        mock_get_client.return_value = client

        runner = CliRunner()
        result = runner.invoke(main, ["federation", "create", "bare-fed"])

        assert result.exit_code == 0
        assert "bare-fed" in result.output
        client._api.post.assert_called_once_with(
            "/federations",
            {"name": "bare-fed", "org_id": "org_test_123"},
        )

    def test_create_requires_api_key(self, monkeypatch, tmp_path):
        monkeypatch.delenv("OCTOMIL_API_KEY", raising=False)
        monkeypatch.setattr(
            "octomil.cli_helpers.os.path.expanduser",
            lambda p: str(tmp_path / p.lstrip("~/")),
        )
        runner = CliRunner()
        result = runner.invoke(main, ["federation", "create", "test"])
        assert result.exit_code != 0


# ---------------------------------------------------------------------------
# federation invite
# ---------------------------------------------------------------------------


class TestFederationInvite:
    @patch("octomil.commands.federation._get_client")
    def test_invite_multiple_orgs(self, mock_get_client, monkeypatch):
        monkeypatch.setenv("OCTOMIL_API_KEY", "test-key")
        client = _mock_client()
        # First call: resolve federation name -> ID
        client._api.get.return_value = [{"id": "fed_001", "name": "my-fed"}]
        # Second call: invite response
        client._api.post.return_value = [
            {"org_id": "org_a", "status": "invited"},
            {"org_id": "org_b", "status": "invited"},
        ]
        mock_get_client.return_value = client

        runner = CliRunner()
        result = runner.invoke(
            main,
            ["federation", "invite", "my-fed", "--org", "org_a", "--org", "org_b"],
        )

        assert result.exit_code == 0
        assert "Invited 2 org(s)" in result.output
        client._api.post.assert_called_once_with(
            "/federations/fed_001/invite",
            {"org_ids": ["org_a", "org_b"]},
        )

    @patch("octomil.commands.federation._get_client")
    def test_invite_single_org(self, mock_get_client, monkeypatch):
        monkeypatch.setenv("OCTOMIL_API_KEY", "test-key")
        client = _mock_client()
        client._api.get.return_value = [{"id": "fed_002", "name": "test-fed"}]
        client._api.post.return_value = [{"org_id": "org_x", "status": "invited"}]
        mock_get_client.return_value = client

        runner = CliRunner()
        result = runner.invoke(
            main,
            ["federation", "invite", "test-fed", "--org", "org_x"],
        )

        assert result.exit_code == 0
        assert "Invited 1 org(s)" in result.output

    def test_invite_requires_org_flag(self, monkeypatch):
        monkeypatch.setenv("OCTOMIL_API_KEY", "test-key")
        runner = CliRunner()
        result = runner.invoke(main, ["federation", "invite", "my-fed"])
        assert result.exit_code != 0

    @patch("octomil.commands.federation._get_client")
    def test_invite_resolves_federation_by_id_fallback(
        self, mock_get_client, monkeypatch
    ):
        """When name lookup returns empty, the raw value is used as ID."""
        monkeypatch.setenv("OCTOMIL_API_KEY", "test-key")
        client = _mock_client()
        # Name lookup returns nothing
        client._api.get.return_value = []
        client._api.post.return_value = {"invited": [{"org_id": "org_z"}]}
        mock_get_client.return_value = client

        runner = CliRunner()
        result = runner.invoke(
            main,
            ["federation", "invite", "fed_raw_id_999", "--org", "org_z"],
        )

        assert result.exit_code == 0
        # Should have called invite with the raw ID
        client._api.post.assert_called_once_with(
            "/federations/fed_raw_id_999/invite",
            {"org_ids": ["org_z"]},
        )


# ---------------------------------------------------------------------------
# federation join
# ---------------------------------------------------------------------------


class TestFederationJoin:
    @patch("octomil.commands.federation._get_client")
    def test_join_success(self, mock_get_client, monkeypatch):
        monkeypatch.setenv("OCTOMIL_API_KEY", "test-key")
        client = _mock_client()
        client._api.get.return_value = [{"id": "fed_join_01", "name": "target-fed"}]
        client._api.post.return_value = {"status": "joined"}
        mock_get_client.return_value = client

        runner = CliRunner()
        result = runner.invoke(main, ["federation", "join", "target-fed"])

        assert result.exit_code == 0
        assert "Joined federation: target-fed" in result.output
        client._api.post.assert_called_once_with(
            "/federations/fed_join_01/join",
            {"org_id": "org_test_123"},
        )

    @patch("octomil.commands.federation._get_client")
    def test_join_with_raw_id(self, mock_get_client, monkeypatch):
        monkeypatch.setenv("OCTOMIL_API_KEY", "test-key")
        client = _mock_client()
        client._api.get.return_value = []  # name not found, fallback to raw ID
        client._api.post.return_value = {"status": "joined"}
        mock_get_client.return_value = client

        runner = CliRunner()
        result = runner.invoke(main, ["federation", "join", "fed_direct_id"])

        assert result.exit_code == 0
        assert "Joined federation: fed_direct_id" in result.output


# ---------------------------------------------------------------------------
# federation list
# ---------------------------------------------------------------------------


class TestFederationList:
    @patch("octomil.commands.federation._get_client")
    def test_list_populated(self, mock_get_client, monkeypatch):
        monkeypatch.setenv("OCTOMIL_API_KEY", "test-key")
        client = _mock_client()
        client._api.get.return_value = [
            {"id": "fed_1", "name": "alpha-fed", "description": "First"},
            {"id": "fed_2", "name": "beta-fed", "description": "Second"},
        ]
        mock_get_client.return_value = client

        runner = CliRunner()
        result = runner.invoke(main, ["federation", "list"])

        assert result.exit_code == 0
        assert "alpha-fed" in result.output
        assert "beta-fed" in result.output
        assert "fed_1" in result.output
        assert "fed_2" in result.output
        assert "First" in result.output
        assert "Second" in result.output

    @patch("octomil.commands.federation._get_client")
    def test_list_empty(self, mock_get_client, monkeypatch):
        monkeypatch.setenv("OCTOMIL_API_KEY", "test-key")
        client = _mock_client()
        client._api.get.return_value = []
        mock_get_client.return_value = client

        runner = CliRunner()
        result = runner.invoke(main, ["federation", "list"])

        assert result.exit_code == 0
        assert "No federations found" in result.output

    @patch("octomil.commands.federation._get_client")
    def test_list_passes_org_id(self, mock_get_client, monkeypatch):
        monkeypatch.setenv("OCTOMIL_API_KEY", "test-key")
        client = _mock_client()
        client._api.get.return_value = []
        mock_get_client.return_value = client

        runner = CliRunner()
        runner.invoke(main, ["federation", "list"])

        client._api.get.assert_called_once_with(
            "/federations",
            params={"org_id": "org_test_123"},
        )


# ---------------------------------------------------------------------------
# federation show
# ---------------------------------------------------------------------------


class TestFederationShow:
    @patch("octomil.commands.federation._get_client")
    def test_show_success(self, mock_get_client, monkeypatch):
        monkeypatch.setenv("OCTOMIL_API_KEY", "test-key")
        client = _mock_client()

        # First get: name resolution
        # Second get: federation details
        call_count = [0]

        def _side_effect(path, params=None):
            call_count[0] += 1
            if call_count[0] == 1:
                # name resolution
                return [{"id": "fed_show_01", "name": "show-fed"}]
            # detail fetch
            return {
                "id": "fed_show_01",
                "name": "show-fed",
                "description": "Detailed desc",
                "created_at": "2026-01-15T10:00:00Z",
                "org_id": "org_test_123",
            }

        client._api.get.side_effect = _side_effect
        mock_get_client.return_value = client

        runner = CliRunner()
        result = runner.invoke(main, ["federation", "show", "show-fed"])

        assert result.exit_code == 0
        assert "show-fed" in result.output
        assert "fed_show_01" in result.output
        assert "Detailed desc" in result.output
        assert "2026-01-15" in result.output

    @patch("octomil.commands.federation._get_client")
    def test_show_with_no_description(self, mock_get_client, monkeypatch):
        monkeypatch.setenv("OCTOMIL_API_KEY", "test-key")
        client = _mock_client()

        call_count = [0]

        def _side_effect(path, params=None):
            call_count[0] += 1
            if call_count[0] == 1:
                return [{"id": "fed_nd", "name": "no-desc"}]
            return {
                "id": "fed_nd",
                "name": "no-desc",
                "description": None,
                "created_at": "2026-02-01T00:00:00Z",
                "org_id": "org_test_123",
            }

        client._api.get.side_effect = _side_effect
        mock_get_client.return_value = client

        runner = CliRunner()
        result = runner.invoke(main, ["federation", "show", "no-desc"])

        assert result.exit_code == 0
        # Description should show '-' for None
        assert "Description: -" in result.output


# ---------------------------------------------------------------------------
# federation members
# ---------------------------------------------------------------------------


class TestFederationMembers:
    @patch("octomil.commands.federation._get_client")
    def test_members_populated(self, mock_get_client, monkeypatch):
        monkeypatch.setenv("OCTOMIL_API_KEY", "test-key")
        client = _mock_client()

        call_count = [0]

        def _side_effect(path, params=None):
            call_count[0] += 1
            if call_count[0] == 1:
                return [{"id": "fed_mem_01", "name": "mem-fed"}]
            return [
                {
                    "org_id": "org_aaa",
                    "status": "active",
                    "joined_at": "2026-01-10T00:00:00Z",
                },
                {"org_id": "org_bbb", "status": "invited", "joined_at": None},
            ]

        client._api.get.side_effect = _side_effect
        mock_get_client.return_value = client

        runner = CliRunner()
        result = runner.invoke(main, ["federation", "members", "mem-fed"])

        assert result.exit_code == 0
        assert "org_aaa" in result.output
        assert "active" in result.output
        assert "org_bbb" in result.output
        assert "invited" in result.output

    @patch("octomil.commands.federation._get_client")
    def test_members_empty(self, mock_get_client, monkeypatch):
        monkeypatch.setenv("OCTOMIL_API_KEY", "test-key")
        client = _mock_client()

        call_count = [0]

        def _side_effect(path, params=None):
            call_count[0] += 1
            if call_count[0] == 1:
                return [{"id": "fed_empty", "name": "empty-fed"}]
            return []

        client._api.get.side_effect = _side_effect
        mock_get_client.return_value = client

        runner = CliRunner()
        result = runner.invoke(main, ["federation", "members", "empty-fed"])

        assert result.exit_code == 0
        assert "No members found" in result.output

    @patch("octomil.commands.federation._get_client")
    def test_members_resolves_federation_name(self, mock_get_client, monkeypatch):
        monkeypatch.setenv("OCTOMIL_API_KEY", "test-key")
        client = _mock_client()

        call_count = [0]

        def _side_effect(path, params=None):
            call_count[0] += 1
            if call_count[0] == 1:
                return [{"id": "fed_resolved", "name": "named-fed"}]
            return [{"org_id": "org_x", "status": "active", "joined_at": "2026-01-01"}]

        client._api.get.side_effect = _side_effect
        mock_get_client.return_value = client

        runner = CliRunner()
        result = runner.invoke(main, ["federation", "members", "named-fed"])

        assert result.exit_code == 0
        # Verify name resolution call
        first_call = client._api.get.call_args_list[0]
        assert first_call[0][0] == "/federations"
        assert first_call[1]["params"]["name"] == "named-fed"
        # Verify members call with resolved ID
        second_call = client._api.get.call_args_list[1]
        assert second_call[0][0] == "/federations/fed_resolved/members"


# ---------------------------------------------------------------------------
# federation share
# ---------------------------------------------------------------------------


class TestFederationShare:
    @patch("octomil.commands.federation._get_client")
    def test_share_model_success(self, mock_get_client, monkeypatch):
        monkeypatch.setenv("OCTOMIL_API_KEY", "test-key")
        client = _mock_client()

        call_count = [0]

        def _get_side_effect(path, params=None):
            call_count[0] += 1
            if "/federations" in path and params and "name" in params:
                return [{"id": "fed_abc", "name": "my-fed"}]
            if "/models" in path:
                return [{"id": "model_123", "name": "radiology-v1"}]
            return []

        client._api.get.side_effect = _get_side_effect
        client._api.post.return_value = {"id": "model_123", "federation_id": "fed_abc"}
        mock_get_client.return_value = client

        runner = CliRunner()
        result = runner.invoke(
            main,
            ["federation", "share", "radiology-v1", "--federation", "my-fed"],
        )

        assert result.exit_code == 0
        assert "shared with federation" in result.output

    @patch("octomil.commands.federation._get_client")
    def test_share_model_not_found(self, mock_get_client, monkeypatch):
        monkeypatch.setenv("OCTOMIL_API_KEY", "test-key")
        client = _mock_client()

        def _get_side_effect(path, params=None):
            if "/federations" in path and params and "name" in params:
                return [{"id": "fed_abc", "name": "my-fed"}]
            if "/models" in path:
                return []  # no models
            return []

        client._api.get.side_effect = _get_side_effect
        mock_get_client.return_value = client

        runner = CliRunner()
        result = runner.invoke(
            main,
            ["federation", "share", "nonexistent", "--federation", "my-fed"],
        )

        assert result.exit_code != 0
        assert "not found" in result.output

    @patch("octomil.commands.federation._get_client")
    def test_share_calls_correct_api(self, mock_get_client, monkeypatch):
        monkeypatch.setenv("OCTOMIL_API_KEY", "test-key")
        client = _mock_client()

        def _get_side_effect(path, params=None):
            if "/federations" in path and params and "name" in params:
                return [{"id": "fed_xyz", "name": "test-fed"}]
            if "/models" in path:
                return [{"id": "model_456", "name": "my-model"}]
            return []

        client._api.get.side_effect = _get_side_effect
        client._api.post.return_value = {}
        mock_get_client.return_value = client

        runner = CliRunner()
        result = runner.invoke(
            main,
            ["federation", "share", "my-model", "--federation", "test-fed"],
        )

        assert result.exit_code == 0
        # Verify the POST call to share the model
        post_calls = client._api.post.call_args_list
        assert len(post_calls) == 1
        assert post_calls[0][0][0] == "/federations/fed_xyz/models"
        assert post_calls[0][0][1] == {"model_id": "model_456"}


# ---------------------------------------------------------------------------
# federation help
# ---------------------------------------------------------------------------


class TestFederationHelp:
    def test_federation_group_help(self):
        runner = CliRunner()
        result = runner.invoke(main, ["federation", "--help"])
        assert result.exit_code == 0
        assert "create" in result.output
        assert "invite" in result.output
        assert "join" in result.output
        assert "list" in result.output
        assert "show" in result.output
        assert "members" in result.output
        assert "share" in result.output

    def test_federation_create_help(self):
        runner = CliRunner()
        result = runner.invoke(main, ["federation", "create", "--help"])
        assert result.exit_code == 0
        assert "--description" in result.output

    def test_federation_share_help(self):
        runner = CliRunner()
        result = runner.invoke(main, ["federation", "share", "--help"])
        assert result.exit_code == 0
        assert "--federation" in result.output



"""Tests for octomil.serve — OpenAI-compatible inference server."""

from __future__ import annotations

import asyncio
import json
from unittest.mock import patch

import pytest
from httpx import ASGITransport, AsyncClient

from octomil.serve import (
    ChatCompletionBody,
    ChatMessage,
    EchoBackend,
    GenerationRequest,
    _detect_backend,
    create_app,
    resolve_model_name,
)

# ---------------------------------------------------------------------------
# resolve_model_name
# ---------------------------------------------------------------------------


class TestResolveModelName:
    def test_full_repo_id_passthrough(self):
        assert resolve_model_name("some-org/some-model", "mlx") == "some-org/some-model"

    def test_full_repo_id_passthrough_gguf(self):
        assert resolve_model_name("user/repo", "gguf") == "user/repo"

    def test_mlx_short_name(self):
        result = resolve_model_name("gemma-1b", "mlx")
        assert result == "mlx-community/REDACTED"

    def test_mlx_unknown_name_raises(self):
        with pytest.raises(ValueError, match="Unknown model 'nonexistent'"):
            resolve_model_name("nonexistent", "mlx")

    def test_gguf_short_name(self):
        result = resolve_model_name("gemma-1b", "gguf")
        assert result == "gemma-1b"  # gguf resolves at download time

    def test_gguf_unknown_name_raises(self):
        with pytest.raises(ValueError, match="Unknown model 'nonexistent'"):
            resolve_model_name("nonexistent", "gguf")

    def test_unknown_backend_unknown_model_raises(self):
        # With the model registry, unknown model names raise even with
        # unknown backends (stricter validation than before).
        with pytest.raises(ValueError, match="Unknown model 'anything'"):
            resolve_model_name("anything", "other")

    def test_unknown_backend_known_model_passthrough(self):
        # Known models still resolve even with unknown backend names
        result = resolve_model_name("gemma-1b", "other")
        assert result == "gemma-1b"


# ---------------------------------------------------------------------------
# EchoBackend
# ---------------------------------------------------------------------------


class TestEchoBackend:
    def setup_method(self):
        self.backend = EchoBackend()
        self.backend.load_model("test-model")

    def test_name(self):
        assert self.backend.name == "echo"

    def test_list_models(self):
        assert self.backend.list_models() == ["test-model"]

    def test_list_models_empty_before_load(self):
        b = EchoBackend()
        assert b.list_models() == []

    def test_generate(self):
        req = GenerationRequest(
            model="test-model",
            messages=[{"role": "user", "content": "hello world"}],
        )
        text, metrics = self.backend.generate(req)
        assert "[echo:test-model]" in text
        assert "hello world" in text
        assert metrics.total_tokens > 0

    def test_generate_empty_messages(self):
        req = GenerationRequest(model="test-model", messages=[])
        text, metrics = self.backend.generate(req)
        assert "[echo:test-model]" in text

    def test_generate_stream(self):
        req = GenerationRequest(
            model="test-model",
            messages=[{"role": "user", "content": "hi"}],
        )
        chunks = asyncio.run(_collect_stream(self.backend.generate_stream(req)))
        assert len(chunks) > 0
        # Last chunk should have finish_reason
        assert chunks[-1].finish_reason == "stop"
        # Reconstructed text should contain the input
        full_text = "".join(c.text for c in chunks)
        assert "hi" in full_text


async def _collect_stream(gen):
    result = []
    async for chunk in gen:
        result.append(chunk)
    return result


# ---------------------------------------------------------------------------
# _detect_backend fallback to echo
# ---------------------------------------------------------------------------


class TestDetectBackend:
    def test_falls_back_to_echo_when_no_backends(self):
        """With no real engines installed, _detect_backend returns echo."""
        from octomil.engines import reset_registry
        from octomil.engines.ollama_engine import OllamaEngine

        reset_registry()
        with (
            patch.dict("sys.modules", {"mlx_lm": None, "llama_cpp": None}),
            patch.object(OllamaEngine, "detect", return_value=False),
        ):
            backend = _detect_backend("some-model")
        assert backend.name == "echo"
        reset_registry()

    def test_echo_for_unknown_model_name(self):
        """Unknown model name still gets a backend (echo fallback)."""
        from octomil.engines import reset_registry
        from octomil.engines.ollama_engine import OllamaEngine

        reset_registry()
        with patch.object(OllamaEngine, "detect", return_value=False):
            backend = _detect_backend("totally-unknown-model")
        # Should return echo since no real engine supports unknown models
        assert backend.name == "echo"
        reset_registry()

    def test_detect_backend_returns_backend(self):
        """_detect_backend returns an InferenceBackend directly."""
        result = _detect_backend("test-model")
        assert hasattr(result, "name")
        assert hasattr(result, "generate")

    def test_detect_backend_with_engine_override(self):
        """_detect_backend with engine_override='echo' returns echo."""
        backend = _detect_backend("test-model", engine_override="echo")
        assert backend.name == "echo"


# ---------------------------------------------------------------------------
# Pydantic models
# ---------------------------------------------------------------------------


class TestPydanticModels:
    def test_chat_message(self):
        msg = ChatMessage(role="user", content="hello")
        assert msg.role == "user"
        assert msg.content == "hello"

    def test_chat_completion_body_defaults(self):
        body = ChatCompletionBody()
        assert body.model == ""
        assert body.messages == []
        assert body.max_tokens == 512
        assert body.temperature == 0.7
        assert body.top_p == 1.0
        assert body.stream is False

    def test_chat_completion_body_with_values(self):
        body = ChatCompletionBody(
            model="test",
            messages=[ChatMessage(role="user", content="hi")],
            max_tokens=100,
            temperature=0.5,
            stream=True,
        )
        assert body.model == "test"
        assert len(body.messages) == 1
        assert body.stream is True


# ---------------------------------------------------------------------------
# FastAPI app (using EchoBackend via create_app)
# ---------------------------------------------------------------------------


@pytest.fixture
def echo_app():
    """Create a FastAPI app with EchoBackend for testing."""
    with patch("octomil.serve._detect_backend") as mock_detect:
        echo = EchoBackend()
        echo.load_model("test-model")
        mock_detect.return_value = echo

        app = create_app("test-model")

        # Trigger lifespan startup manually
        async def _trigger_lifespan():
            ctx = app.router.lifespan_context(app)
            await ctx.__aenter__()

        asyncio.run(_trigger_lifespan())

    return app


@pytest.mark.asyncio
async def test_health_endpoint(echo_app):
    transport = ASGITransport(app=echo_app)
    async with AsyncClient(transport=transport, base_url="http://test") as client:
        resp = await client.get("/health")
    assert resp.status_code == 200
    data = resp.json()
    assert data["status"] == "ok"
    assert data["model"] == "test-model"
    assert data["backend"] == "echo"


@pytest.mark.asyncio
async def test_list_models(echo_app):
    transport = ASGITransport(app=echo_app)
    async with AsyncClient(transport=transport, base_url="http://test") as client:
        resp = await client.get("/v1/models")
    assert resp.status_code == 200
    data = resp.json()
    assert data["object"] == "list"
    assert len(data["data"]) == 1
    assert data["data"][0]["id"] == "test-model"
    assert data["data"][0]["owned_by"] == "octomil"


@pytest.mark.asyncio
async def test_chat_completions_non_streaming(echo_app):
    transport = ASGITransport(app=echo_app)
    async with AsyncClient(transport=transport, base_url="http://test") as client:
        resp = await client.post(
            "/v1/chat/completions",
            json={
                "model": "test-model",
                "messages": [{"role": "user", "content": "hello"}],
                "max_tokens": 50,
            },
        )
    assert resp.status_code == 200
    data = resp.json()
    assert data["object"] == "chat.completion"
    assert len(data["choices"]) == 1
    assert data["choices"][0]["message"]["role"] == "assistant"
    assert "hello" in data["choices"][0]["message"]["content"]
    assert data["choices"][0]["finish_reason"] == "stop"
    assert "usage" in data


@pytest.mark.asyncio
async def test_chat_completions_streaming(echo_app):
    transport = ASGITransport(app=echo_app)
    async with AsyncClient(transport=transport, base_url="http://test") as client:
        resp = await client.post(
            "/v1/chat/completions",
            json={
                "model": "test-model",
                "messages": [{"role": "user", "content": "hi"}],
                "stream": True,
            },
        )
    assert resp.status_code == 200

    # Parse SSE events
    lines = resp.text.strip().split("\n\n")
    events = []
    for line in lines:
        if line.startswith("data: ") and line != "data: [DONE]":
            events.append(json.loads(line[6:]))

    assert len(events) > 0
    # Each event should be a chat.completion.chunk
    for evt in events:
        assert evt["object"] == "chat.completion.chunk"
        assert len(evt["choices"]) == 1

    # Last real event should have finish_reason
    last_event = events[-1]
    assert last_event["choices"][0]["finish_reason"] == "stop"

    # Final line should be [DONE]
    assert lines[-1] == "data: [DONE]"


@pytest.mark.asyncio
async def test_chat_completions_empty_messages(echo_app):
    transport = ASGITransport(app=echo_app)
    async with AsyncClient(transport=transport, base_url="http://test") as client:
        resp = await client.post(
            "/v1/chat/completions",
            json={"messages": []},
        )
    assert resp.status_code == 200


@pytest.mark.asyncio
async def test_chat_completions_model_default(echo_app):
    """When model is omitted, should use the server's model name."""
    transport = ASGITransport(app=echo_app)
    async with AsyncClient(transport=transport, base_url="http://test") as client:
        resp = await client.post(
            "/v1/chat/completions",
            json={
                "messages": [{"role": "user", "content": "test"}],
            },
        )
    assert resp.status_code == 200
    data = resp.json()
    assert data["model"] == "test-model"



"""Tests for cloud streaming inference (SSE parsing + OctomilClient integration)."""

from __future__ import annotations

import unittest
from unittest.mock import MagicMock, patch

from octomil.streaming import (
    StreamToken,
    _build_payload,
    _parse_sse_line,
    _parse_sse_lines,
    stream_inference,
)


# ------------------------------------------------------------------
# StreamToken dataclass
# ------------------------------------------------------------------


class StreamTokenTests(unittest.TestCase):
    def test_defaults(self):
        tok = StreamToken(token="hello", done=False)
        self.assertEqual(tok.token, "hello")
        self.assertFalse(tok.done)
        self.assertIsNone(tok.provider)
        self.assertIsNone(tok.latency_ms)
        self.assertIsNone(tok.session_id)

    def test_all_fields(self):
        tok = StreamToken(
            token="world",
            done=True,
            provider="ollama",
            latency_ms=42.5,
            session_id="abc-123",
        )
        self.assertEqual(tok.token, "world")
        self.assertTrue(tok.done)
        self.assertEqual(tok.provider, "ollama")
        self.assertAlmostEqual(tok.latency_ms, 42.5)
        self.assertEqual(tok.session_id, "abc-123")


# ------------------------------------------------------------------
# _build_payload
# ------------------------------------------------------------------


class BuildPayloadTests(unittest.TestCase):
    def test_string_input(self):
        p = _build_payload("phi-4-mini", "Hello", None)
        self.assertEqual(p["model_id"], "phi-4-mini")
        self.assertEqual(p["input_data"], "Hello")
        self.assertNotIn("messages", p)
        self.assertNotIn("parameters", p)

    def test_messages_input(self):
        msgs = [{"role": "user", "content": "Hi"}]
        p = _build_payload("phi-4-mini", msgs, None)
        self.assertEqual(p["messages"], msgs)
        self.assertNotIn("input_data", p)

    def test_with_parameters(self):
        params = {"temperature": 0.7, "max_tokens": 512}
        p = _build_payload("phi-4-mini", "test", params)
        self.assertEqual(p["parameters"], params)

    def test_empty_parameters_omitted(self):
        p = _build_payload("phi-4-mini", "test", {})
        self.assertNotIn("parameters", p)


# ------------------------------------------------------------------
# _parse_sse_line
# ------------------------------------------------------------------


class ParseSSELineTests(unittest.TestCase):
    def test_normal_token(self):
        line = 'data: {"token": "The", "done": false, "provider": "ollama"}'
        tok = _parse_sse_line(line)
        self.assertIsNotNone(tok)
        self.assertEqual(tok.token, "The")
        self.assertFalse(tok.done)
        self.assertEqual(tok.provider, "ollama")

    def test_done_token(self):
        line = 'data: {"done": true, "latency_ms": 1234.5, "session_id": "abc-123"}'
        tok = _parse_sse_line(line)
        self.assertIsNotNone(tok)
        self.assertTrue(tok.done)
        self.assertEqual(tok.token, "")
        self.assertAlmostEqual(tok.latency_ms, 1234.5)
        self.assertEqual(tok.session_id, "abc-123")

    def test_empty_line_returns_none(self):
        self.assertIsNone(_parse_sse_line(""))
        self.assertIsNone(_parse_sse_line("   "))

    def test_non_data_line_returns_none(self):
        self.assertIsNone(_parse_sse_line("event: message"))
        self.assertIsNone(_parse_sse_line("id: 1"))
        self.assertIsNone(_parse_sse_line(": comment"))

    def test_empty_data_returns_none(self):
        self.assertIsNone(_parse_sse_line("data:"))
        self.assertIsNone(_parse_sse_line("data:   "))

    def test_invalid_json_returns_none(self):
        self.assertIsNone(_parse_sse_line("data: not-json"))

    def test_leading_trailing_whitespace(self):
        line = '  data: {"token": "x", "done": false}  '
        tok = _parse_sse_line(line)
        self.assertIsNotNone(tok)
        self.assertEqual(tok.token, "x")


# ------------------------------------------------------------------
# _parse_sse_lines
# ------------------------------------------------------------------


class ParseSSELinesTests(unittest.TestCase):
    def test_multiple_lines(self):
        lines = [
            'data: {"token": "The", "done": false}',
            'data: {"token": " answer", "done": false}',
            "",
            'data: {"done": true, "latency_ms": 100.0, "session_id": "s1"}',
        ]
        tokens = list(_parse_sse_lines(iter(lines)))
        self.assertEqual(len(tokens), 3)
        self.assertEqual(tokens[0].token, "The")
        self.assertEqual(tokens[1].token, " answer")
        self.assertTrue(tokens[2].done)

    def test_empty_stream(self):
        tokens = list(_parse_sse_lines(iter([])))
        self.assertEqual(tokens, [])


# ------------------------------------------------------------------
# stream_inference (mocked HTTP)
# ------------------------------------------------------------------


class StreamInferenceTests(unittest.TestCase):
    def test_stream_inference_yields_tokens(self):
        sse_lines = [
            'data: {"token": "Hello", "done": false, "provider": "ollama"}',
            'data: {"token": " world", "done": false, "provider": "ollama"}',
            'data: {"done": true, "latency_ms": 50.0, "session_id": "sess-1"}',
        ]

        mock_response = MagicMock()
        mock_response.raise_for_status = MagicMock()
        mock_response.iter_lines.return_value = iter(sse_lines)
        mock_response.__enter__ = MagicMock(return_value=mock_response)
        mock_response.__exit__ = MagicMock(return_value=False)

        mock_client = MagicMock()
        mock_client.stream.return_value = mock_response
        mock_client.__enter__ = MagicMock(return_value=mock_client)
        mock_client.__exit__ = MagicMock(return_value=False)

        with patch("octomil.streaming.httpx.Client", return_value=mock_client):
            tokens = list(
                stream_inference(
                    server_url="https://api.octomil.com/api/v1",
                    api_key="test-key",
                    model_id="phi-4-mini",
                    input_data="Hello",
                )
            )

        self.assertEqual(len(tokens), 3)
        self.assertEqual(tokens[0].token, "Hello")
        self.assertFalse(tokens[0].done)
        self.assertEqual(tokens[1].token, " world")
        self.assertTrue(tokens[2].done)
        self.assertAlmostEqual(tokens[2].latency_ms, 50.0)

    def test_stream_inference_sends_correct_request(self):
        mock_response = MagicMock()
        mock_response.raise_for_status = MagicMock()
        mock_response.iter_lines.return_value = iter([])
        mock_response.__enter__ = MagicMock(return_value=mock_response)
        mock_response.__exit__ = MagicMock(return_value=False)

        mock_client = MagicMock()
        mock_client.stream.return_value = mock_response
        mock_client.__enter__ = MagicMock(return_value=mock_client)
        mock_client.__exit__ = MagicMock(return_value=False)

        with patch("octomil.streaming.httpx.Client", return_value=mock_client):
            list(
                stream_inference(
                    server_url="https://api.octomil.com/api/v1/",
                    api_key="my-key",
                    model_id="phi-4-mini",
                    input_data="test prompt",
                    parameters={"temperature": 0.5},
                )
            )

        mock_client.stream.assert_called_once()
        call_args = mock_client.stream.call_args
        self.assertEqual(call_args[0][0], "POST")
        self.assertEqual(
            call_args[0][1], "https://api.octomil.com/api/v1/inference/stream"
        )
        self.assertEqual(call_args[1]["json"]["model_id"], "phi-4-mini")
        self.assertEqual(call_args[1]["json"]["input_data"], "test prompt")
        self.assertEqual(call_args[1]["json"]["parameters"]["temperature"], 0.5)
        self.assertEqual(call_args[1]["headers"]["Authorization"], "Bearer my-key")

    def test_stream_inference_with_messages(self):
        mock_response = MagicMock()
        mock_response.raise_for_status = MagicMock()
        mock_response.iter_lines.return_value = iter([])
        mock_response.__enter__ = MagicMock(return_value=mock_response)
        mock_response.__exit__ = MagicMock(return_value=False)

        mock_client = MagicMock()
        mock_client.stream.return_value = mock_response
        mock_client.__enter__ = MagicMock(return_value=mock_client)
        mock_client.__exit__ = MagicMock(return_value=False)

        msgs = [{"role": "user", "content": "Hi"}]
        with patch("octomil.streaming.httpx.Client", return_value=mock_client):
            list(
                stream_inference(
                    server_url="https://api.octomil.com/api/v1",
                    api_key="key",
                    model_id="phi-4-mini",
                    input_data=msgs,
                )
            )

        call_json = mock_client.stream.call_args[1]["json"]
        self.assertEqual(call_json["messages"], msgs)
        self.assertNotIn("input_data", call_json)


# ------------------------------------------------------------------
# stream_inference_async (mocked HTTP)
# ------------------------------------------------------------------


class StreamInferenceAsyncTests(unittest.IsolatedAsyncioTestCase):
    async def test_stream_inference_async_yields_tokens(self):
        from unittest.mock import AsyncMock

        from octomil.streaming import stream_inference_async

        sse_lines = [
            'data: {"token": "Async", "done": false}',
            'data: {"token": " test", "done": false}',
            'data: {"done": true, "session_id": "s2"}',
        ]

        async def mock_aiter_lines():
            for line in sse_lines:
                yield line

        mock_response = MagicMock()
        mock_response.raise_for_status = MagicMock()
        mock_response.aiter_lines = mock_aiter_lines

        mock_stream_ctx = MagicMock()
        mock_stream_ctx.__aenter__ = AsyncMock(return_value=mock_response)
        mock_stream_ctx.__aexit__ = AsyncMock(return_value=False)

        mock_client = MagicMock()
        mock_client.stream.return_value = mock_stream_ctx

        mock_client_ctx = MagicMock()
        mock_client_ctx.__aenter__ = AsyncMock(return_value=mock_client)
        mock_client_ctx.__aexit__ = AsyncMock(return_value=False)

        with patch("octomil.streaming.httpx.AsyncClient", return_value=mock_client_ctx):
            tokens = []
            async for tok in stream_inference_async(
                server_url="https://api.octomil.com/api/v1",
                api_key="key",
                model_id="phi-4-mini",
                input_data="Hello",
            ):
                tokens.append(tok)

        self.assertEqual(len(tokens), 3)
        self.assertEqual(tokens[0].token, "Async")
        self.assertTrue(tokens[2].done)


# ------------------------------------------------------------------
# OctomilClient.stream_predict integration (mocked)
# ------------------------------------------------------------------


class ClientStreamPredictTests(unittest.TestCase):
    def test_stream_predict_delegates_to_stream_inference(self):
        from octomil.client import OctomilClient

        expected_tokens = [
            StreamToken(token="Hi", done=False),
            StreamToken(token="", done=True, session_id="s1"),
        ]

        with patch(
            "octomil.streaming.stream_inference", return_value=iter(expected_tokens)
        ) as mock_fn:
            client = OctomilClient(
                api_key="test-key", api_base="https://api.test.com/api/v1"
            )
            tokens = list(
                client.stream_predict(
                    "phi-4-mini",
                    "Hello",
                    parameters={"temperature": 0.7},
                )
            )

        self.assertEqual(len(tokens), 2)
        self.assertEqual(tokens[0].token, "Hi")
        mock_fn.assert_called_once_with(
            server_url="https://api.test.com/api/v1",
            api_key="test-key",
            model_id="phi-4-mini",
            input_data="Hello",
            parameters={"temperature": 0.7},
            timeout=120.0,
        )


class ClientStreamPredictAsyncTests(unittest.IsolatedAsyncioTestCase):
    async def test_stream_predict_async_delegates(self):
        from octomil.client import OctomilClient

        expected_tokens = [
            StreamToken(token="Async", done=False),
            StreamToken(token="", done=True),
        ]

        async def fake_stream_async(**kwargs):
            for tok in expected_tokens:
                yield tok

        with patch(
            "octomil.streaming.stream_inference_async",
            side_effect=lambda **kwargs: fake_stream_async(**kwargs),
        ):
            client = OctomilClient(
                api_key="test-key", api_base="https://api.test.com/api/v1"
            )
            tokens = []
            async for tok in client.stream_predict_async("phi-4-mini", "Hello"):
                tokens.append(tok)

        self.assertEqual(len(tokens), 2)
        self.assertEqual(tokens[0].token, "Async")


if __name__ == "__main__":
    unittest.main()



"""Tests for octomil.enterprise — Enterprise API client and CLI commands."""

from __future__ import annotations

import json
from unittest.mock import MagicMock, patch

import pytest
from click.testing import CliRunner

from octomil.cli import main
from octomil.enterprise import (
    COMPLIANCE_PRESETS,
    EnterpriseClient,
    EnterpriseClientError,
    load_config,
    save_config,
)


# ---------------------------------------------------------------------------
# Compliance presets
# ---------------------------------------------------------------------------


class TestCompliancePresets:
    def test_hipaa_preset_has_required_fields(self):
        hipaa = COMPLIANCE_PRESETS["hipaa"]
        assert hipaa["hipaa_mode"] is True
        assert hipaa["audit_retention_days"] == 2190
        assert hipaa["require_mfa_for_admin"] is True
        assert hipaa["policy_profile"] == "regulated"
        assert hipaa["session_duration_hours"] == 8
        assert hipaa["reauth_interval_minutes"] == 30

    def test_gdpr_preset_has_required_fields(self):
        gdpr = COMPLIANCE_PRESETS["gdpr"]
        assert gdpr["audit_retention_days"] == 1825
        assert gdpr["require_mfa_for_admin"] is True
        assert gdpr["policy_profile"] == "balanced"

    def test_pci_preset_has_required_fields(self):
        pci = COMPLIANCE_PRESETS["pci"]
        assert pci["require_mfa_for_admin"] is True
        assert pci["mfa_required"] is True
        assert pci["session_duration_hours"] == 8
        assert pci["require_admin_approval"] is True
        assert pci["policy_profile"] == "regulated"

    def test_soc2_preset_has_required_fields(self):
        soc2 = COMPLIANCE_PRESETS["soc2"]
        assert soc2["audit_retention_days"] == 365
        assert soc2["require_mfa_for_admin"] is True
        assert soc2["require_admin_approval"] is True
        assert soc2["require_model_approval"] is True
        assert soc2["auto_rollback_enabled"] is True
        assert soc2["policy_profile"] == "balanced"

    def test_all_presets_have_policy_profile(self):
        for name, preset in COMPLIANCE_PRESETS.items():
            assert "policy_profile" in preset, f"{name} missing policy_profile"
            assert preset["policy_profile"] in ("startup", "balanced", "regulated")


# ---------------------------------------------------------------------------
# Config file helpers
# ---------------------------------------------------------------------------


class TestConfigHelpers:
    def test_save_and_load_config(self, tmp_path, monkeypatch):
        config_file = tmp_path / ".octomil" / "config.json"
        monkeypatch.setattr("octomil.enterprise._config_path", lambda: config_file)

        data = {"org_id": "test-org-123", "org_name": "Test Corp", "region": "us"}
        save_config(data)

        assert config_file.exists()
        loaded = load_config()
        assert loaded["org_id"] == "test-org-123"
        assert loaded["org_name"] == "Test Corp"
        assert loaded["region"] == "us"

    def test_load_config_returns_empty_when_missing(self, tmp_path, monkeypatch):
        config_file = tmp_path / ".octomil" / "config.json"
        monkeypatch.setattr("octomil.enterprise._config_path", lambda: config_file)
        assert load_config() == {}

    def test_load_config_returns_empty_on_bad_json(self, tmp_path, monkeypatch):
        config_file = tmp_path / ".octomil" / "config.json"
        config_file.parent.mkdir(parents=True, exist_ok=True)
        config_file.write_text("not json at all{{{")
        monkeypatch.setattr("octomil.enterprise._config_path", lambda: config_file)
        assert load_config() == {}

    def test_save_config_creates_parent_dirs(self, tmp_path, monkeypatch):
        config_file = tmp_path / "deep" / "nested" / ".octomil" / "config.json"
        monkeypatch.setattr("octomil.enterprise._config_path", lambda: config_file)
        save_config({"org_id": "x"})
        assert config_file.exists()

    def test_config_file_has_restricted_permissions(self, tmp_path, monkeypatch):
        config_file = tmp_path / ".octomil" / "config.json"
        monkeypatch.setattr("octomil.enterprise._config_path", lambda: config_file)
        save_config({"org_id": "x"})
        mode = config_file.stat().st_mode & 0o777
        assert mode == 0o600


# ---------------------------------------------------------------------------
# EnterpriseClient
# ---------------------------------------------------------------------------


class TestEnterpriseClient:
    def _mock_client(self, monkeypatch):
        """Return an EnterpriseClient with a mocked httpx.Client."""
        mock_http = MagicMock()
        client = EnterpriseClient.__new__(EnterpriseClient)
        client._http = mock_http
        return client, mock_http

    def test_create_org(self, monkeypatch):
        client, mock_http = self._mock_client(monkeypatch)
        resp = MagicMock()
        resp.status_code = 200
        resp.json.return_value = {"org_id": "acme-corp", "workspace_type": "enterprise"}
        mock_http.put.return_value = resp

        result = client.create_org("Acme Corp", region="us")
        assert result["org_id"] == "acme-corp"
        mock_http.put.assert_called_once()
        call_args = mock_http.put.call_args
        assert "/onboarding/" in call_args[0][0]
        assert call_args[1]["json"]["workspace_name"] == "Acme Corp"

    def test_set_policy_profile(self, monkeypatch):
        client, mock_http = self._mock_client(monkeypatch)
        resp = MagicMock()
        resp.status_code = 200
        resp.json.return_value = {"policy_profile": "regulated"}
        mock_http.put.return_value = resp

        result = client.set_policy_profile("org-1", "regulated")
        assert result["policy_profile"] == "regulated"
        mock_http.put.assert_called_once_with(
            "/onboarding/org-1/policy",
            json={"policy_profile": "regulated"},
        )

    def test_set_compliance_hipaa(self, monkeypatch):
        client, mock_http = self._mock_client(monkeypatch)
        resp = MagicMock()
        resp.status_code = 200
        resp.json.return_value = {"hipaa_mode": True}
        mock_http.put.return_value = resp

        result = client.set_compliance("org-1", "hipaa")
        # Should have called put twice: once for policy profile, once for settings
        assert mock_http.put.call_count == 2

    def test_set_compliance_unknown_raises(self, monkeypatch):
        client, mock_http = self._mock_client(monkeypatch)
        with pytest.raises(EnterpriseClientError, match="Unknown compliance preset"):
            client.set_compliance("org-1", "unknown_preset")

    def test_invite_member(self, monkeypatch):
        client, mock_http = self._mock_client(monkeypatch)
        resp = MagicMock()
        resp.status_code = 201
        resp.json.return_value = {"email": "alice@acme.com", "role": "admin"}
        mock_http.post.return_value = resp

        result = client.invite_member("org-1", "alice@acme.com", role="admin")
        assert result["email"] == "alice@acme.com"
        assert result["role"] == "admin"
        mock_http.post.assert_called_once()

    def test_list_members(self, monkeypatch):
        client, mock_http = self._mock_client(monkeypatch)
        resp = MagicMock()
        resp.status_code = 200
        resp.json.return_value = [
            {"email": "alice@acme.com", "role": "admin"},
            {"email": "bob@acme.com", "role": "member"},
        ]
        mock_http.get.return_value = resp

        members = client.list_members("org-1")
        assert len(members) == 2
        mock_http.get.assert_called_once_with("/orgs/org-1/members")

    def test_create_api_key(self, monkeypatch):
        client, mock_http = self._mock_client(monkeypatch)
        resp = MagicMock()
        resp.status_code = 201
        resp.json.return_value = {
            "api_key": "edg_test1234567890",
            "prefix": "edg_test1234",
            "name": "deploy-key",
        }
        mock_http.post.return_value = resp

        result = client.create_api_key(
            "org-1", "deploy-key", scopes={"devices": "write"}
        )
        assert "api_key" in result
        assert result["name"] == "deploy-key"

    def test_list_api_keys(self, monkeypatch):
        client, mock_http = self._mock_client(monkeypatch)
        resp = MagicMock()
        resp.status_code = 200
        resp.json.return_value = [
            {"name": "key1", "prefix": "edg_abc"},
            {"name": "key2", "prefix": "edg_def"},
        ]
        mock_http.get.return_value = resp

        keys = client.list_api_keys("org-1")
        assert len(keys) == 2

    def test_revoke_api_key(self, monkeypatch):
        client, mock_http = self._mock_client(monkeypatch)
        resp = MagicMock()
        resp.status_code = 200
        resp.json.return_value = {
            "name": "old-key",
            "revoked_at": "2026-01-01T00:00:00Z",
        }
        mock_http.post.return_value = resp

        result = client.revoke_api_key("key-123")
        assert result["revoked_at"] is not None
        mock_http.post.assert_called_once_with("/api-keys/key-123/revoke")

    def test_get_settings(self, monkeypatch):
        client, mock_http = self._mock_client(monkeypatch)
        resp = MagicMock()
        resp.status_code = 200
        resp.json.return_value = {
            "org_id": "org-1",
            "audit_retention_days": 90,
            "require_mfa_for_admin": False,
        }
        mock_http.get.return_value = resp

        result = client.get_settings("org-1")
        assert result["audit_retention_days"] == 90

    def test_update_settings(self, monkeypatch):
        client, mock_http = self._mock_client(monkeypatch)
        resp = MagicMock()
        resp.status_code = 200
        resp.json.return_value = {"require_mfa_for_admin": True}
        mock_http.put.return_value = resp

        result = client.update_settings("org-1", require_mfa_for_admin=True)
        assert result["require_mfa_for_admin"] is True
        mock_http.put.assert_called_once_with(
            "/settings/org-1",
            json={"require_mfa_for_admin": True},
        )

    def test_get_onboarding_status(self, monkeypatch):
        client, mock_http = self._mock_client(monkeypatch)
        resp = MagicMock()
        resp.status_code = 200
        resp.json.return_value = {"org_id": "org-1", "current_step": "sdk_setup"}
        mock_http.get.return_value = resp

        result = client.get_onboarding_status("org-1")
        assert result["current_step"] == "sdk_setup"

    def test_complete_onboarding(self, monkeypatch):
        client, mock_http = self._mock_client(monkeypatch)
        resp = MagicMock()
        resp.status_code = 200
        resp.json.return_value = {
            "current_step": "complete",
            "onboarding_complete": True,
        }
        mock_http.post.return_value = resp

        result = client.complete_onboarding("org-1")
        assert result["onboarding_complete"] is True

    def test_raise_for_status_on_error(self, monkeypatch):
        client, mock_http = self._mock_client(monkeypatch)
        resp = MagicMock()
        resp.status_code = 403
        resp.json.return_value = {"detail": "Forbidden"}
        resp.text = '{"detail": "Forbidden"}'
        mock_http.get.return_value = resp

        with pytest.raises(EnterpriseClientError, match="403"):
            client.get_settings("org-1")

    def test_raise_for_status_non_json_error(self, monkeypatch):
        client, mock_http = self._mock_client(monkeypatch)
        resp = MagicMock()
        resp.status_code = 500
        resp.json.side_effect = ValueError("not json")
        resp.text = "Internal Server Error"
        mock_http.get.return_value = resp

        with pytest.raises(EnterpriseClientError, match="Internal Server Error"):
            client.get_settings("org-1")


# ---------------------------------------------------------------------------
# CLI: octomil init
# ---------------------------------------------------------------------------


class TestInitCommand:
    @patch("octomil.commands.enterprise._get_api_key", return_value="test-key")
    @patch("octomil.enterprise.EnterpriseClient")
    @patch("octomil.enterprise.save_config")
    @patch("octomil.enterprise.load_config", return_value={})
    def test_init_creates_org(self, mock_load, mock_save, MockClient, mock_key):
        mock_instance = MockClient.return_value
        mock_instance.create_org.return_value = {"org_id": "acme-corp"}

        runner = CliRunner()
        result = runner.invoke(main, ["init", "Acme Corp", "--region", "us"])

        assert result.exit_code == 0
        assert "acme-corp" in result.output
        mock_instance.create_org.assert_called_once_with(
            "Acme Corp", region="us", workspace_type="enterprise"
        )
        mock_save.assert_called_once()
        saved_config = mock_save.call_args[0][0]
        assert saved_config["org_id"] == "acme-corp"
        assert saved_config["org_name"] == "Acme Corp"

    @patch("octomil.commands.enterprise._get_api_key", return_value="test-key")
    @patch("octomil.enterprise.EnterpriseClient")
    @patch("octomil.enterprise.save_config")
    @patch("octomil.enterprise.load_config", return_value={})
    def test_init_with_compliance(self, mock_load, mock_save, MockClient, mock_key):
        mock_instance = MockClient.return_value
        mock_instance.create_org.return_value = {"org_id": "acme-corp"}
        mock_instance.set_compliance.return_value = {}

        runner = CliRunner()
        result = runner.invoke(
            main, ["init", "Acme Corp", "--compliance", "hipaa", "--region", "eu"]
        )

        assert result.exit_code == 0
        assert "HIPAA" in result.output
        mock_instance.set_compliance.assert_called_once_with("acme-corp", "hipaa")
        saved_config = mock_save.call_args[0][0]
        assert saved_config["compliance"] == "hipaa"
        assert saved_config["region"] == "eu"

    @patch("octomil.commands.enterprise._get_api_key", return_value="")
    def test_init_requires_api_key(self, mock_key):
        runner = CliRunner()
        result = runner.invoke(main, ["init", "Acme Corp"])
        assert result.exit_code != 0
        assert "No API key" in result.output

    def test_init_help(self):
        runner = CliRunner()
        result = runner.invoke(main, ["init", "--help"])
        assert result.exit_code == 0
        assert "compliance" in result.output
        assert "region" in result.output


# ---------------------------------------------------------------------------
# CLI: octomil team
# ---------------------------------------------------------------------------


class TestTeamCommands:
    def test_team_help(self):
        runner = CliRunner()
        result = runner.invoke(main, ["team", "--help"])
        assert result.exit_code == 0
        assert "add" in result.output
        assert "list" in result.output
        assert "set-policy" in result.output

    @patch("octomil.commands.enterprise._require_org_id", return_value="org-123")
    @patch("octomil.commands.enterprise._get_enterprise_client")
    def test_team_add(self, mock_get_client, mock_org_id):
        mock_client = MagicMock()
        mock_client.invite_member.return_value = {
            "email": "alice@acme.com",
            "role": "admin",
        }
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(
            main, ["team", "add", "alice@acme.com", "--role", "admin"]
        )
        assert result.exit_code == 0
        assert "alice@acme.com" in result.output
        assert "admin" in result.output
        mock_client.invite_member.assert_called_once_with(
            "org-123", "alice@acme.com", role="admin", name=None
        )

    @patch("octomil.commands.enterprise._require_org_id", return_value="org-123")
    @patch("octomil.commands.enterprise._get_enterprise_client")
    def test_team_list(self, mock_get_client, mock_org_id):
        mock_client = MagicMock()
        mock_client.list_members.return_value = [
            {"email": "alice@acme.com", "role": "admin", "name": "Alice"},
            {"email": "bob@acme.com", "role": "member", "name": "Bob"},
        ]
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(main, ["team", "list"])
        assert result.exit_code == 0
        assert "alice@acme.com" in result.output
        assert "bob@acme.com" in result.output
        assert "admin" in result.output
        assert "member" in result.output
        assert "2 member(s)" in result.output

    @patch("octomil.commands.enterprise._require_org_id", return_value="org-123")
    @patch("octomil.commands.enterprise._get_enterprise_client")
    def test_team_list_empty(self, mock_get_client, mock_org_id):
        mock_client = MagicMock()
        mock_client.list_members.return_value = []
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(main, ["team", "list"])
        assert result.exit_code == 0
        assert "No team members found" in result.output

    @patch("octomil.commands.enterprise._require_org_id", return_value="org-123")
    @patch("octomil.commands.enterprise._get_enterprise_client")
    def test_team_set_policy_mfa(self, mock_get_client, mock_org_id):
        mock_client = MagicMock()
        mock_client.update_settings.return_value = {}
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(
            main, ["team", "set-policy", "--require-mfa", "--session-hours", "8"]
        )
        assert result.exit_code == 0
        mock_client.update_settings.assert_called_once_with(
            "org-123",
            require_mfa_for_admin=True,
            session_duration_hours=8,
        )

    @patch("octomil.commands.enterprise._require_org_id", return_value="org-123")
    @patch("octomil.commands.enterprise._get_enterprise_client")
    def test_team_set_policy_auto_rollback(self, mock_get_client, mock_org_id):
        mock_client = MagicMock()
        mock_client.update_settings.return_value = {}
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(
            main,
            ["team", "set-policy", "--auto-rollback", "--audit-retention-days", "365"],
        )
        assert result.exit_code == 0
        mock_client.update_settings.assert_called_once_with(
            "org-123",
            auto_rollback_enabled=True,
            audit_retention_days=365,
        )

    @patch("octomil.commands.enterprise._require_org_id", return_value="org-123")
    @patch("octomil.commands.enterprise._get_enterprise_client")
    def test_team_set_policy_no_changes(self, mock_get_client, mock_org_id):
        runner = CliRunner()
        result = runner.invoke(main, ["team", "set-policy"])
        assert result.exit_code == 0
        assert "No policy changes" in result.output


# ---------------------------------------------------------------------------
# CLI: octomil keys
# ---------------------------------------------------------------------------


class TestKeysCommands:
    def test_keys_help(self):
        runner = CliRunner()
        result = runner.invoke(main, ["keys", "--help"])
        assert result.exit_code == 0
        assert "create" in result.output
        assert "list" in result.output
        assert "revoke" in result.output

    @patch("octomil.commands.enterprise._require_org_id", return_value="org-123")
    @patch("octomil.commands.enterprise._get_enterprise_client")
    def test_keys_create(self, mock_get_client, mock_org_id):
        mock_client = MagicMock()
        mock_client.create_api_key.return_value = {
            "api_key": "edg_test1234567890abcdef",
            "prefix": "edg_test1234",
            "name": "deploy-key",
        }
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(
            main,
            [
                "keys",
                "create",
                "deploy-key",
                "--scope",
                "devices:write",
                "--scope",
                "models:read",
            ],
        )
        assert result.exit_code == 0
        assert "deploy-key" in result.output
        assert "edg_test1234567890abcdef" in result.output
        assert "Save this key" in result.output

        # Verify scopes were passed correctly
        call_kwargs = mock_client.create_api_key.call_args
        assert call_kwargs[0] == ("org-123", "deploy-key")
        scopes = call_kwargs[1]["scopes"]
        assert scopes["devices"] == "write"
        assert scopes["models"] == "read"

    @patch("octomil.commands.enterprise._require_org_id", return_value="org-123")
    @patch("octomil.commands.enterprise._get_enterprise_client")
    def test_keys_create_no_scopes(self, mock_get_client, mock_org_id):
        mock_client = MagicMock()
        mock_client.create_api_key.return_value = {
            "api_key": "edg_abc123",
            "prefix": "edg_abc1",
            "name": "admin-key",
        }
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(main, ["keys", "create", "admin-key"])
        assert result.exit_code == 0
        call_kwargs = mock_client.create_api_key.call_args
        assert call_kwargs[1]["scopes"] is None

    @patch("octomil.commands.enterprise._require_org_id", return_value="org-123")
    @patch("octomil.commands.enterprise._get_enterprise_client")
    def test_keys_list(self, mock_get_client, mock_org_id):
        mock_client = MagicMock()
        mock_client.list_api_keys.return_value = [
            {
                "name": "deploy-key",
                "prefix": "edg_abc12345",
                "created_at": "2026-01-15T10:00:00Z",
                "revoked_at": None,
            },
            {
                "name": "old-key",
                "prefix": "edg_def67890",
                "created_at": "2025-06-01T10:00:00Z",
                "revoked_at": "2025-12-01T10:00:00Z",
            },
        ]
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(main, ["keys", "list"])
        assert result.exit_code == 0
        assert "deploy-key" in result.output
        assert "old-key" in result.output
        assert "2 key(s)" in result.output

    @patch("octomil.commands.enterprise._require_org_id", return_value="org-123")
    @patch("octomil.commands.enterprise._get_enterprise_client")
    def test_keys_list_empty(self, mock_get_client, mock_org_id):
        mock_client = MagicMock()
        mock_client.list_api_keys.return_value = []
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(main, ["keys", "list"])
        assert result.exit_code == 0
        assert "No API keys found" in result.output

    @patch("octomil.commands.enterprise._get_enterprise_client")
    def test_keys_revoke(self, mock_get_client):
        mock_client = MagicMock()
        mock_client.revoke_api_key.return_value = {
            "name": "old-key",
            "prefix": "edg_abc",
            "revoked_at": "2026-02-19T00:00:00Z",
        }
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(main, ["keys", "revoke", "key-abc-123", "--yes"])
        assert result.exit_code == 0
        assert "Revoked" in result.output
        assert "old-key" in result.output
        mock_client.revoke_api_key.assert_called_once_with("key-abc-123")


# ---------------------------------------------------------------------------
# CLI: octomil org
# ---------------------------------------------------------------------------


class TestOrgCommand:
    @patch("octomil.enterprise.get_org_id", return_value="")
    def test_org_no_config(self, mock_org):
        runner = CliRunner()
        result = runner.invoke(main, ["org"])
        assert result.exit_code == 0
        assert "No organization configured" in result.output

    @patch("octomil.commands.enterprise._get_api_key", return_value="test-key")
    @patch("octomil.commands.enterprise._get_enterprise_client")
    @patch("octomil.enterprise.get_org_id", return_value="org-123")
    @patch(
        "octomil.enterprise.load_config",
        return_value={
            "org_name": "Test Corp",
            "region": "us",
            "compliance": "soc2",
        },
    )
    def test_org_shows_info(self, mock_load, mock_org_id, mock_get_client, mock_key):
        mock_client = MagicMock()
        mock_client.get_settings.return_value = {
            "audit_retention_days": 365,
            "require_mfa_for_admin": True,
            "require_admin_approval": True,
            "require_model_approval": True,
            "auto_rollback_enabled": True,
            "session_duration_hours": 12,
            "reauth_interval_minutes": 60,
        }
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(main, ["org"])
        assert result.exit_code == 0
        assert "Test Corp" in result.output
        assert "org-123" in result.output
        assert "SOC2" in result.output
        assert "365" in result.output


# ---------------------------------------------------------------------------
# get_org_id
# ---------------------------------------------------------------------------


class TestGetOrgId:
    def test_from_env(self, monkeypatch):
        from octomil.enterprise import get_org_id

        monkeypatch.setenv("OCTOMIL_ORG_ID", "env-org-42")
        assert get_org_id() == "env-org-42"

    def test_from_config(self, monkeypatch, tmp_path):
        from octomil.enterprise import get_org_id

        monkeypatch.delenv("OCTOMIL_ORG_ID", raising=False)
        config_file = tmp_path / ".octomil" / "config.json"
        config_file.parent.mkdir(parents=True, exist_ok=True)
        config_file.write_text(json.dumps({"org_id": "config-org-99"}))
        monkeypatch.setattr("octomil.enterprise._config_path", lambda: config_file)

        assert get_org_id() == "config-org-99"

    def test_empty_when_no_env_no_config(self, monkeypatch, tmp_path):
        from octomil.enterprise import get_org_id

        monkeypatch.delenv("OCTOMIL_ORG_ID", raising=False)
        config_file = tmp_path / ".octomil" / "config.json"
        monkeypatch.setattr("octomil.enterprise._config_path", lambda: config_file)

        assert get_org_id() == ""
